import sys
import tempfile
import zipfile
import os
import json
import io

if sys.version_info.minor > 6:
    python_36 = False
    from contextlib import nullcontext
else:
    python_36 = True
    from contextlib import suppress as nullcontext
from xrnn.data_handler import DataHandler
from xrnn import layer_utils
from xrnn.model import Model
from xrnn import activations
from xrnn import optimizers
from xrnn import losses
from xrnn import config
from xrnn import layers
from xrnn import ops
import re
import pytest


BATCH_SIZE = 64
INPUT_SHAPE = (BATCH_SIZE, 28, 28, 3)
IMAGE_BATCH = ops.zeros(INPUT_SHAPE)
LABELS = ops.ones((BATCH_SIZE,))
X = ops.random.standard_normal((200, *INPUT_SHAPE[1:]))
Y = ops.ones(len(X))


def one_hot_encoded(n_classes):
    integers = ops.random.integers(0, n_classes, (BATCH_SIZE,))
    one_hot = ops.zeros((BATCH_SIZE, n_classes))
    one_hot[ops.arange(len(integers)), integers] = 1
    return one_hot


@pytest.fixture
def cnn():
    m = Model()
    m.add(layers.Conv2D(16, 3, 2))
    m.add(layers.BatchNormalization())
    m.add(activations.LeakyReLU())
    m.add(layers.Dropout(0.25))
    m.add(layers.AvgPooling2D(2, 2, 'same'))
    m.add(layers.Flatten())
    m.add(layers.Dense(10))
    m.add(activations.Softmax())
    m.set('adam', 'categorical_crossentropy')
    return m


class TestModel:

    @pytest.mark.parametrize(
        "opt, loss, raises",
        [
            ('adam', 'mse', nullcontext()),
            ('Adam', 'MSE', nullcontext()),
            ('rmsprop', 'rmsprop', pytest.raises(ValueError)),
            ('adam', 'categorical_crossentropy', nullcontext()),
            (optimizers.Adam(), 'mse', nullcontext()),
            ('sgd', losses.BinaryCrossentropy(), nullcontext()),
            ('adagrad', 'binary_crossentropy', nullcontext()),
            ('adamm', losses.CategoricalCrossentropy(), pytest.raises(ValueError))
        ]
    )
    def test_set(self, opt, loss, raises):
        with raises:
            model = Model()
            model.set(opt, loss)
            if isinstance(opt, str):
                opt_str = {
                    'adam': optimizers.Adam, 'sgd': optimizers.SGD,
                    'rmsprop': optimizers.RMSprop, 'adagrad': optimizers.Adagrad}
                opt = opt_str[opt.lower()]()
            if isinstance(loss, str):
                loss_str = {
                    'mse': losses.MSE,
                    'categorical_crossentropy': losses.CategoricalCrossentropy,
                    'binary_crossentropy': losses.BinaryCrossentropy}
                loss = loss_str[loss.lower()]()
            assert isinstance(model.optimizer, type(opt))
            assert isinstance(model.loss, type(loss))
            assert model.loss.trainable_layers == model.trainable_layers
            # Test set after add
            model = Model()
            model.add(layers.Dense(10))
            model.set('adam', 'mse')
            assert model.loss.trainable_layers == model.trainable_layers

    def test_add(self):
        m = Model()
        m.set('adam', 'mse')
        m.add(layers.Dense(10))
        m.add(layers.Flatten())
        assert len(m.layers) == 2
        assert len(m.trainable_layers) == 1
        assert len(m.loss.trainable_layers) == 1
        # Test add after build
        m.build((BATCH_SIZE, 100))
        m.add(layers.Dense(16))
        assert m.layers[-1].built
        assert m.layers[-1].weights.shape == (10, 16)
        # Test adding a built layer to a built model with wrong shape.
        d = layers.Dense(16)
        d.build((BATCH_SIZE, 100))
        with pytest.raises(ValueError):
            m.add(d)
        # Test adding a built layer to a built model with correct shape.
        d = layers.Dense(32)
        d.build((BATCH_SIZE, 16))
        m.add(d)
        assert m.output_shape == (BATCH_SIZE, 32)

    def test_build(self):
        m = Model()
        m.add(layers.Dense(16))
        m.add(activations.ReLU())
        m.add(layers.Dense(128))
        m.build((BATCH_SIZE, 100))
        assert m.layers[0].built
        assert m.layers[0].weights.shape == (100, 16)
        assert m.layers[0].weights.max() <= ops.sqrt(6 / 100)  # Make sure 'he' initialization method is used
        assert m.layers[-1].built
        assert m.layers[-1].weights.shape == (16, 128)
        assert m.layers[-1].weights.max() <= ops.sqrt(6 / (16 + 128))  # Make sure 'xavier' method is used.

        # Test adding an already built layer then calling model.build().
        m = Model()
        layer = layers.Dense(16)
        layer.build((BATCH_SIZE, 100))
        m.add(layer)
        # Test wrong input shape
        with pytest.raises(ValueError):
            m.build((BATCH_SIZE, 128))
        # Test correct input shape
        m.build((BATCH_SIZE, 100))
        assert m.output_shape == (BATCH_SIZE, 16)

        # Test wrong input shape for layers that don't have a build method but except input_shape to be in a certain way
        m = Model()
        m.add(layers.MaxPooling2D(3))
        with pytest.raises(ValueError):
            m.build((BATCH_SIZE, 100))  # inputs must be 4D
        m.build(INPUT_SHAPE)
        assert m.layers[0].input_shape is not None

    def test_output_shape(self):
        config.set_image_data_format('channels-last')
        m = Model()
        with pytest.raises(ValueError):
            _ = m.output_shape
        m.add(layers.Conv2D(32, 3, 2))
        m.add(layers.MaxPooling2D(2, 2, 'same'))
        m.build((BATCH_SIZE, 64, 64, 3))
        assert m.layers[-1].input_shape == m.layers[0].output_shape == (BATCH_SIZE, 31, 31, 32)
        assert m.output_shape == m.layers[-1].output_shape == (BATCH_SIZE, 16, 16, 32)
        m.add(layers.Flatten())
        assert m.layers[-1].input_shape == (BATCH_SIZE, 16, 16, 32)
        assert m.output_shape == m.layers[-1].output_shape == (BATCH_SIZE, 16 * 16 * 32)

    def test_forward(self, cnn):
        cnn.build(INPUT_SHAPE)
        with pytest.raises(ValueError):
            cnn.forward(ops.zeros((BATCH_SIZE, 7, 7, 3)), True)
        output = cnn.forward(IMAGE_BATCH, False)
        assert output.shape == cnn.output_shape
        assert output.shape == (BATCH_SIZE, 10)
        assert output.dtype == 'float32'
        # 10 classes, softmax activation function, each class has a 0.1 probability when the model isn't trained.
        assert output.max() == pytest.approx(0.1)
        assert cnn.layers[0].inputs is None
        assert cnn.layers[1].training
        cnn.forward(IMAGE_BATCH, True)
        assert cnn.layers[0].inputs is IMAGE_BATCH
        config.set_default_dtype('float64')
        output = cnn.forward(IMAGE_BATCH[:32], False)
        assert cnn.layers[0].inputs is not IMAGE_BATCH  # The datatype changed, so it shouldn't be the same array.
        assert output.dtype == config.DTYPE

    def test_call(self, cnn):
        output = cnn(IMAGE_BATCH)
        assert ops.array_equal(output, cnn.forward(IMAGE_BATCH))

    @pytest.mark.parametrize('labels', [LABELS, one_hot_encoded(10)])
    def test_backward_fast_path(self, labels, cnn):
        # backward has a fast path when the last layer is Softmax and loss is categorical crossentropy.
        y_pred = cnn.forward(IMAGE_BATCH, False).astype('f4')
        with pytest.raises(AttributeError):
            cnn.backward(labels, y_pred)
        cnn.forward(IMAGE_BATCH, True)
        cnn.layers[0].training = False
        config.set_default_dtype('f4')
        output = cnn.backward(labels, y_pred)
        assert output.shape == cnn.input_shape
        assert output.dtype == 'float32'

    @pytest.mark.parametrize('loss', ['mse', 'binary_crossentropy', 'categorical_crossentropy'])
    def test_backward_slow_path(self, loss, cnn):
        # Remove the last (Softmax) layer so the slower path is used when the loss is categorical crossentropy,
        # otherwise, it wouldn't matter.
        if loss == 'categorical_crossentropy':
            cnn.layers.pop()
        cnn.set(cnn.optimizer, loss)
        y_pred = cnn.forward(IMAGE_BATCH)
        assert cnn.backward(y_pred.copy(), y_pred).shape == cnn.input_shape

    @pytest.mark.parametrize('batch_size', [16, 32, BATCH_SIZE, 128])
    def test_update_progressbar(self, batch_size, cnn, capfd):
        cnn.train(X, Y, batch_size, 1)
        out, err = capfd.readouterr()
        # From 1 to -1, because the 0th element is the epoch counter, and the last element is the same progressbar with
        # different formatting.
        updates = out.split('\r')[1:-1]
        for i, update in enumerate(updates):
            assert update.split()[1] == f'{i + 1}/{len(updates)}'  # number of steps
            assert len(update.split()[2]) == 32  # The progressbar part
            assert update.split('-')[1].strip() == f'{((i + 1) / len(updates)):.0%}'
        assert 'Took' in out.split('\r')[-1]
        # Test that only two new lines were printed, one after the epoch counter and one after the progressbar.
        assert out.count('\n') == 2
        cnn.train(X, Y, batch_size, 2, validation_split=0.1)
        out, err = capfd.readouterr()
        assert out.count('val_loss') == 2
        assert out.count('Took') == 2
        # 6 newlines, 2 for the first epoch and progressbar, 1 line between them and the second epoch, 2 for the 2nd
        # epoch, 1 for print(Training took).
        assert out.count('\n') == 6

    @pytest.mark.parametrize(
        "val_split, val_data",
        [
            (0.1, None),
            (0.1, (IMAGE_BATCH, ops.ones(len(IMAGE_BATCH)))),
            (0., [IMAGE_BATCH, ops.ones(len(IMAGE_BATCH))]),
            (None, (IMAGE_BATCH, ops.ones(len(IMAGE_BATCH)))),
            (None, DataHandler(IMAGE_BATCH, ops.ones(len(IMAGE_BATCH)))),
            (None, None),
            (0., None)
        ]
    )
    def test_train_validation_data(self, val_split, val_data, cnn, capfd):
        cnn.train(X, Y, BATCH_SIZE, validation_split=val_split, validation_data=val_data)
        out, err = capfd.readouterr()
        if any([val_split, val_data]):
            assert 'val_loss' in out
        else:
            assert 'val_loss' not in out
        # Test validation_split when x is a DataHandler object and not an array.
        if val_split and not val_data:
            with pytest.raises(TypeError):
                cnn.train(DataHandler(X, Y), validation_split=val_split, validation_data=val_data)

    @pytest.mark.parametrize(
        "epochs, print_every, valid_freq, should_print",
        [
            (1, 1, 1, True),
            (1, 2, 0, False),
            (2, 0, 0, False),
            (5, False, False, False),
            (1, None, None, False),
            (2, 2, 0, True),
            (2, 1, 3, True),
            (6, 2, 3, True),
        ]
    )
    def test_train_verbose(self, epochs, print_every, valid_freq, should_print, cnn, capfd):
        cnn.train(X, Y, BATCH_SIZE, epochs, print_every=print_every, validation_split=0.1, validation_freq=valid_freq)
        out, err = capfd.readouterr()
        assert bool(out) == should_print
        if should_print:
            # Test that is printed every 'print_every'th epoch.
            assert out.count('Epoch') == epochs // print_every
            assert bool('Training took' in out) == bool(epochs >= 2)  # It should print that only if epochs>=2
            if valid_freq and valid_freq * print_every <= epochs:
                assert 'val_loss' in out
            else:
                assert 'val_loss' not in out

    @pytest.mark.parametrize("steps_per_epoch", [1, 2, 3, 4, 5])
    def test_train_steps_per_epoch(self, steps_per_epoch, cnn, capfd):
        cnn.train(X, Y, BATCH_SIZE, steps_per_epoch=steps_per_epoch)
        out, err = capfd.readouterr()
        assert cnn.optimizer.iterations == min(steps_per_epoch, len(X) // BATCH_SIZE + 1)
        # Test that the progressbar print the right number of steps.
        assert out.splitlines()[2].split('/')[1][0] == str(min(steps_per_epoch, len(X) // BATCH_SIZE + 1))

    def test_train_optimizer_updates(self, cnn):
        m = Model()
        # Test calling train before `set`.
        with pytest.raises(ValueError):
            m.train(X, Y)
        cnn.layers[0].training = False
        cnn.build(INPUT_SHAPE)
        before_weights = [cnn.layers[0].weights.copy(), cnn.layers[1].weights.copy()]
        cnn.train(X, Y, BATCH_SIZE)
        # Test that the optimizer didn't update non-trainable layers.
        assert ops.array_equal(cnn.layers[0].weights, before_weights[0])
        # Test that the optimizer updates trainable layers.
        assert not ops.array_equal(cnn.layers[1].weights, before_weights[1])

    def test_evaluate(self, cnn):
        loss, acc = cnn.evaluate(IMAGE_BATCH, LABELS).values()
        assert cnn.layers[0].inputs is None
        assert loss > 2
        assert acc == 0

    def test_inference(self, cnn):
        output = cnn.inference(IMAGE_BATCH)
        assert output.shape == (INPUT_SHAPE[0], *cnn.output_shape[1:])
        output = cnn.inference(IMAGE_BATCH, 6)
        assert output.shape == (INPUT_SHAPE[0], *cnn.output_shape[1:])

    def test_mem_usage(self, cnn):
        with pytest.raises(ValueError):
            cnn.mem_usage()
        cnn.build(INPUT_SHAPE)
        config.set_default_dtype('f4')
        mem_usage = cnn.mem_usage()
        assert mem_usage == cnn.mem_usage(BATCH_SIZE)
        assert mem_usage < cnn.mem_usage(BATCH_SIZE * 2)
        config.set_default_dtype(float)
        assert cnn.mem_usage() == mem_usage * 2
        config.set_default_dtype('f4')
        # Test optimizers other than adam mem usage.
        cnn.set('adagrad', 'categorical_crossentropy')
        assert cnn.mem_usage() < mem_usage

    def test_summary(self, capfd, cnn):
        # Test calling `summary()` with a big batch size that isn't multiple of 2.
        m = Model()
        m.add(layers.MaxPooling2D(2))
        m.build((1025, 28, 28, 3))
        with pytest.warns(UserWarning):
            m.summary()
        # call `readouterr()` to clear it for the next print
        out, err = capfd.readouterr()
        # Test calling `summary()` before `build()`.
        with pytest.raises(ValueError):
            cnn.summary()
        cnn.build(INPUT_SHAPE)
        cnn.summary()
        out, err = capfd.readouterr()
        lines = [line for line in out.splitlines() if line]
        # Got the re pattern from ChatGPT; first time I used something that came out of it, the memes are actually True.
        first_layer_line = re.match(r'(\S+)\s+(\d+)\s+(\([\d\s,]+\))\s+(\([\d\s,]+\))', lines[3])
        assert first_layer_line[1] == cnn.layers[0].name
        assert first_layer_line[2] == str(cnn.layers[0].weights.size + cnn.layers[0].biases.size)
        assert first_layer_line[3] == str(cnn.input_shape)
        assert first_layer_line[4] == str(cnn.layers[0].output_shape)
        assert '-' in lines[0]
        assert '-' in lines[-1]
        assert '=' in lines[2]
        assert '=' in lines[-8]
        assert len(lines[0]) == len(lines[-1]) == len(lines[2]) == len(lines[-8])
        assert len([line for line in out.split('=' * len(lines[2]))[1].splitlines() if line]) == len(cnn.layers)
        assert lines[-2].split(':')[-1].strip() == layer_utils.to_readable_unit_converter(cnn.mem_usage())
        non_trainable_params_before = lines[-6].split(':')[-1].strip()
        trainable_params_before = lines[-7].split(':')[-1].strip()
        cnn.layers[0].training = False
        cnn.summary()
        out, err = capfd.readouterr()
        lines = [line for line in out.splitlines() if line]
        # Test that setting a layer training state changes the parameter count correctly
        assert lines[-6].split(':')[-1].strip() > non_trainable_params_before
        assert lines[-7].split(':')[-1].strip() < trainable_params_before

    def test_get_params(self, cnn):
        m = Model()
        # Test get params on an empty model.
        with pytest.raises(ValueError):
            m.get_params()
        # Test get params on a model with no trainable layers.
        m.add(layers.MaxPooling2D((2, 2)))
        with pytest.raises(ValueError):
            m.get_params()
        # Test get params before build.
        with pytest.raises(ValueError):
            cnn.get_params()
        cnn.build(INPUT_SHAPE)
        model_parameters = cnn.get_params()
        assert len(model_parameters) == 3

    def test_set_params(self, cnn):
        with pytest.raises(ValueError):
            cnn.set_params([[] for _ in range(len(cnn.trainable_layers))])
        cnn.build(INPUT_SHAPE)
        model_parameters = cnn.get_params()
        # Test that setting params actually works and changes the model's layers' parameters.
        cnn.layers[0].set_params([ops.ones_like(cnn.layers[0].weights)])
        cnn.set_params(model_parameters)
        assert ops.array_equal(cnn.layers[0].weights, model_parameters[0][0])
        with pytest.raises(IndexError):
            cnn.set_params(model_parameters[:-1])
        with pytest.raises(IndexError):
            model_parameters.append([])
            cnn.set_params(model_parameters)

    def test_save_params(self, cnn):
        cnn.build(INPUT_SHAPE)
        temp_file_path = os.path.join(tempfile.gettempdir(), 'temp_params.npz')
        cnn.save_params(temp_file_path)
        try:
            with ops.load(temp_file_path) as model_params_dict:
                assert len(model_params_dict) == 8
                assert len([key for key in model_params_dict if "Conv2D" in key]) == 2
                assert len([key for key in model_params_dict if "BatchNormalization" in key]) == 4
                assert len([key for key in model_params_dict if "Dense" in key]) == 2
                assert len([key for key in model_params_dict if key.endswith('0')]) == 3
                assert all([model_params_dict[key].dtype == config.DTYPE] for key in model_params_dict)
        finally:  # To ensure the file gets deleted.
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_load_params(self, cnn):
        temp_file_path = os.path.join(tempfile.gettempdir(), 'temp_params.npz')
        try:
            cnn.build(INPUT_SHAPE)
            cnn.save_params(temp_file_path)
            cnn.load_params(temp_file_path)
            # Test less params saved than layers.
            cnn.add(layers.Dense(16))
            with pytest.raises(IndexError):
                cnn.load_params(temp_file_path)
            # Test more params saved than layers.
            cnn.layers.pop()
            cnn.layers.pop()
            with pytest.raises(IndexError):
                cnn.load_params(temp_file_path)
        finally:  # To ensure the file gets deleted.
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_save(self, cnn):
        temp_file_path = os.path.join(tempfile.gettempdir(), 'temp_model.xmodel')
        try:
            with pytest.raises(ValueError):
                cnn.save(temp_file_path)
            cnn.build(INPUT_SHAPE)
            cnn.save(temp_file_path)
            assert os.path.exists(temp_file_path)
            with zipfile.ZipFile(temp_file_path) as saved_model:
                assert saved_model.namelist() == ['model_config.json', 'model_params.npz']
                with saved_model.open(
                        'model_config.json') as config_file, saved_model.open('model_params.npz') as model_params_file:
                    model_config = json.load(config_file)
                    # The following is to convert all lists in dict to tuples for all dicts in a list.
                    # This is a single line, but I split to not violate PEP8 (probably the least of your concerns
                    # when looking at it).
                    # Writing a loop that accomplishes the same is easier and more readable, but I like this.
                    model_config['layers_configs'] = [
                        dict(
                            zip(
                                a.keys(),
                                list(
                                    map(
                                        lambda x: tuple(x) if isinstance(x, list) and len(x) > 1 else x,
                                        a.values()
                                    )
                                )
                            )
                        ) for a in model_config['layers_configs']
                    ]
                    # Test that the model's config was saved correctly.
                    assert model_config == cnn.get_config()
                    if python_36:
                        model_params_file = io.BytesIO(model_params_file.read())
                    loaded_params = ops.load(model_params_file)
                    assert len(loaded_params) == 8  # 2 for conv, 2 for dense, 4 for batch norm.
        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_load(self, cnn):
        # `Model.load()` depends only on `zipfile.ZipFile()`, `json.load` (both hopefully tested thoroughly),
        # `Model.from_config()`, `Model.load_params()` (tested separately), so not much to test here apart from testing
        # that the pieces work together.
        temp_file_path = os.path.join(tempfile.gettempdir(), 'temp_model.xmodel')
        try:
            cnn.build(INPUT_SHAPE)
            cnn.save(temp_file_path)
            new_cnn = cnn.load(temp_file_path)
            assert len(new_cnn.layers) == len(cnn.layers)
            assert new_cnn.built
            assert new_cnn.output_shape == cnn.output_shape
            # Tests that the layers are constructed with the same arguments as the orginal one.
            assert new_cnn.get_config() == cnn.get_config()
        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_get_config(self):
        model = Model()
        model_config = model.get_config()
        assert list(model_config.keys()) == ['layers_configs', 'opt_config', 'loss']
        assert len(model_config['layers_configs']) == 0
        assert model_config['opt_config'] is None
        assert model_config['loss'] is None
        model.add(layers.Dense(16))
        model.set('adam', 'mse')
        model_config = model.get_config()
        assert model_config['layers_configs'][0] == model.layers[0].get_config()
        assert model_config['opt_config'] == optimizers.Adam().get_config()
        assert model_config['loss'] == 'MSE'

    @pytest.mark.parametrize('build', [False, True])
    def test_from_config(self, build, cnn):
        if build:
            cnn.build(INPUT_SHAPE)
        model_config = cnn.get_config()
        new_cnn = Model.from_config(model_config, build)
        assert len(new_cnn.layers) == len(cnn.layers)
        assert new_cnn.optimizer.get_config() == cnn.optimizer.get_config()
        assert type(new_cnn.loss) is type(cnn.loss)
        assert new_cnn.built is build
        for new_layer, old_layer in zip(new_cnn.layers, cnn.layers):
            assert new_layer.get_config() == old_layer.get_config()
        if build:
            assert new_cnn.output_shape == cnn.output_shape
            new_cnn.forward(IMAGE_BATCH)
        # Test when the optimizer and loss aren't set.
        model_config['opt_config'] = None
        new_cnn = Model.from_config(model_config)
        assert new_cnn.optimizer is new_cnn.loss is None
        # Test from_config with a custom layer.
        model_config['layers_configs'].append({"type": "MyCustomLayer"})
        with pytest.raises(NotImplementedError):
            Model.from_config(model_config)



"""Defines metric classes that can be used to calculate the model's performance against said metric."""
from typing import Union

from xrnn import losses
from xrnn import ops


class Accuracy:

    def __init__(
            self, loss: Union[ops.config.Literal[
                'mse', 'binary_crossentropy', 'categorical_crossentropy'], losses.Loss]) -> None:
        """
        Measures the model accumulated accuracy.

        Parameters
        ----------
        loss: str, Loss
            Loss function class or a string representing the loss function used, available options are 'mse',
            'binary_crossentropy', 'categorical_crossentropy'.
        """
        self.accumulated_acc = 0
        self.accumulated_count = 0

        if isinstance(loss, str):
            if loss.lower() not in ('mse', 'binary_crossentropy', 'categorical_crossentropy'):
                raise ValueError('`loss` must be one of "mse", "binary_crossentropy", "categorical_crossentropy".')
            loss = loss.lower()

        if isinstance(loss, losses.CategoricalCrossentropy) or loss == 'categorical_crossentropy':
            self.acc_function = self.categorical_accuracy
        if isinstance(loss, losses.BinaryCrossentropy) or loss == 'binary_crossentropy':
            self.acc_function = lambda y_true, y_pred: (y_pred > 0.5).astype(int) == y_true
        if isinstance(loss, losses.MeanSquaredError) or loss == 'mse':
            self.acc_function = lambda y_true, y_pred: ops.absolute(y_true - y_pred) < (ops.std(y_true) / 250)

    def reset_count(self) -> None:
        """Resets the accumulated accuracy and step count to start over again, called at the start of each epoch."""
        self.accumulated_acc = 0
        self.accumulated_count = 0

    @staticmethod
    def categorical_accuracy(y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        """
        Calculates the categorical accuracy. Used to assess a classification model accuracy.

        .. note::
           This function calculates accuracy sample-wise, meaning it calculates accuracy for each sample in the batch,
           for calculating accuracy as a number [0, 1], use `calculate`.
        """
        if y_true.ndim == 2:
            if y_true.shape[1] == 1:  # If the second dimension is 1, we can get rid of it.
                y_true = ops.squeeze(y_true)
            else:  # if labels are one-hot encoded, convert them to a sparse array
                y_true = ops.argmax(y_true, axis=1)
        return ops.argmax(y_pred, axis=1) == y_true  # noqa, both are numpy arrays so the result is a bool array.

    def calculate(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> float:
        """
        Calculates the model accuracy. There are three different types of accuracy, classification, regression and
        binary accuracy, this method decides which one to use based on the loss function.

        Notes
        -----
        This method calculates accumulated accuracy and not step accuracy.
        """
        comparisons = self.acc_function(y_true, y_pred)
        self.accumulated_acc += ops.sum(comparisons)
        self.accumulated_count += comparisons.size  # noqa, it's a numpy array Pycharm!
        return self.accumulated_acc / self.accumulated_count

    def __call__(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> float:
        """Same as `calculate`. Implemented just to have the same API as other classes in this package where you can get
        the result by calling the object and not the `calculate/forward` method directly."""
        return self.calculate(y_true, y_pred)



# pylint: skip-file
# flake8: noqa
"""
This module has all the names imported by `import numpy` just using its namespace (ops.ones for e.g.), it extends the
functionality of some numpy functions such as `zeros` to use the dtype defined in `config.py` by default.
 * Note that user passed arguments have the priority over the default ones.
 * Methods in `numpy.random` are available under `ops._random`.
 * `ops.random` is and instance of `~numpy.random.Generator`.
 This way the code in this package can call `ops.random` functions as it would call `numpy.random` functions but with
 the added benefit of having a separate RNG specific to it, so for example, calling `numpy.random.seed(0)` won't affect
 the results generated from `ops.random` functions. Setting a seed is achieved just like numpy by calling `random.seed`.
 """
from typing import Union

from numpy import *

from xrnn import config

np_methods_to_warp = (zeros, ones, eye)
_random = random
dtype_hint = Union[str, type, dtype]
shape_hint = Union[int, tuple]


def zeros(shape: shape_hint, dtype: dtype_hint = None, order: str = 'C') -> ndarray:
    if not dtype:
        dtype = config.DTYPE
    return np_methods_to_warp[0](shape, dtype, order)


def ones(shape: shape_hint, dtype: dtype_hint = None, order: str = 'C') -> ndarray:
    if not dtype:
        dtype = config.DTYPE
    return np_methods_to_warp[1](shape, dtype, order)


def eye(N: int, M: int = None, k: int = 0, dtype: dtype_hint = None, order: str = 'C') -> ndarray:
    if not dtype:
        dtype = config.DTYPE
    return np_methods_to_warp[2](N, M, k, dtype, order)


zeros.__doc__ = np_methods_to_warp[0].__doc__
ones.__doc__ = np_methods_to_warp[1].__doc__
eye.__doc__ = np_methods_to_warp[2].__doc__


class _Random(_random.Generator):

    def __init__(self, seed=None):
        super().__init__(_random.PCG64(seed))

    def standard_normal(self, shape: shape_hint = None, dtype: dtype_hint = None, out: ndarray = None) -> ndarray:
        if not dtype:
            dtype = config.DTYPE
        return super().standard_normal(shape, dtype, out)

    def uniform(self, low: float = 0., high: float = 1., shape: shape_hint = None, dtype: dtype_hint = None) -> ndarray:
        if not dtype:
            dtype = config.DTYPE
        return ascontiguousarray(super().uniform(low, high, shape), dtype)

    def binomial(self, n: int, p: Union[float, ndarray], shape: shape_hint = None, dtype: dtype_hint = None) -> ndarray:
        if not dtype:
            dtype = config.DTYPE
        return ascontiguousarray(super().binomial(n, p, shape), dtype)

    def seed(self, seed=None):
        """
        Sets a new seed for the generator resulting in new generated random numbers.

        Parameters
        ----------
        seed: {None, int, array_like[ints], SeedSequence}, optional
            A seed to initialize the `BitGenerator`. If None, then fresh,
            unpredictable entropy will be pulled from the OS. If an ``int`` or
            ``array_like[ints]`` is passed, then it will be passed to
            `SeedSequence` to derive the initial `BitGenerator` state. One may also
            pass in a `SeedSequence` instance. *Note*: To pick a new random seed automatically pass None.
        """
        super().__init__(_random.PCG64(seed))


random = _Random()



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
from xrnn import activations
from xrnn import layers
from xrnn import config
from xrnn import ops
import pytest

nhwc_shape = (128, 32, 32, 3)
nchw_shape = (128, 3, 32, 32)


@pytest.fixture
def dense():
    d = layers.Dense(16)
    d.build(100)
    return d


@pytest.fixture
def spatial_layer():
    def _create_spatial_layer(w=3, s=2, p=None):
        if not p:
            p = 'same'
        return layers.SpatialLayer(w, s, p)

    return _create_spatial_layer


class TestLayer:

    @pytest.mark.parametrize(
        "dtype, expected",
        [
            (ops.float32, 'float32'),
            (ops.float64, 'float64'),
            ('d', 'float64'),
            ('f8', 'float64'),
            (float, 'float64'),
            ('single', 'float32'),
            (ops.ones(10, 'f'), 'float32')
        ]
    )
    def test_dtype(self, dense, dtype, expected):
        dense.dtype = dtype
        assert dense.dtype == expected
        assert dense.weights.dtype == expected
        assert dense.biases.dtype == expected

    def test_output_shape(self):
        layer = layers.Layer()
        with pytest.raises(ValueError):
            assert layer.output_shape
        layer.input_shape = (128, 32, 32, 3)
        assert layer.output_shape == (128, 32, 32, 3)

    def test_call(self):
        layer = layers.Dense(16)
        inputs = ops.ones((64, 100)).tolist()
        output = layer(inputs)
        assert layer.built
        assert layer.input_shape == (64, 100)
        assert isinstance(layer.inputs, ops.ndarray)
        assert layer.inputs.dtype == config.DTYPE
        assert output.dtype == config.DTYPE
        with pytest.raises(ValueError):
            layer(ops.ones(nhwc_shape))
        with pytest.raises(ValueError):
            layer(ops.ones((64, 128)))
        layer(ops.ones((128, 100)))
        assert layer.input_shape == (128, 100)

    @pytest.mark.parametrize(
        "method, activation, expect",
        [
            ('standard_normal', None, nullcontext()),
            ('auto', 'tanh', nullcontext()),
            ('', 'relu', pytest.raises(ValueError)),
            ('xavier', 'relu', nullcontext()),
            ('he', 'sigmoid', nullcontext()),
            ('auto', 'fail', pytest.raises(ValueError))
        ]
    )
    def test_get_initialization_function(self, method, activation, expect):
        with expect:
            layers.Layer().get_initialization_function(method, activation)

    def test_get_params(self):
        layer = layers.Dense(16)
        # Test get_params before build
        with pytest.raises(ValueError):
            layer.get_params()
        layer.build(100)
        initial_weights, initial_biases = layer.weights.copy(), layer.biases.copy()
        weights, biases = layer.get_params(False)
        assert weights.shape == (100, 16)
        assert biases.shape == (16,)
        # Test that setting copy to False doesn't copy the arrays.
        assert id(weights) == id(layer.weights)
        assert id(biases) == id(layer.biases)
        weights, biases = layer.get_params()
        assert id(weights) != id(layer.weights)
        assert id(biases) != id(layer.biases)

    def test_set_params(self):
        layer = layers.Dense(16)
        # Test set before build.
        with pytest.raises(ValueError):
            layer.set_params([])
        layer.build(100)
        layer.set_params(layer.get_params())
        # Test empty parameters list
        with pytest.raises(ValueError):
            layer.set_params([])
        # Test wrong isn't a list type.
        with pytest.raises(TypeError):
            layer.set_params(layer.get_params()[0])
        # Test params elements aren't numpy arrays.
        with pytest.raises(TypeError):
            layer.set_params([None])
        # Test that setting only weights works.
        b_weights, b_biases = layer.get_params(False)
        layer.set_params([layer.get_params()[0]])
        assert id(layer.weights) != id(b_weights)
        assert id(layer.biases) == id(b_biases)
        layer.set_params(layer.get_params())
        assert id(layer.weights) != id(b_weights)
        assert id(layer.biases) != id(b_biases)
        with pytest.raises(ValueError):
            layer.set_params([ops.ones(10)])
        with pytest.raises(ValueError):
            layer.set_params([layer.weights, ops.ones(10)])

    def test_get_config(self, dense):
        layer = layers.Layer()
        layer_config = layer.get_config()
        assert len(layer_config) == 3
        assert layer_config['type'] == 'Layer'
        assert layer_config['dtype'] == config.DTYPE
        assert layer_config['training'] is True
        assert dense.get_config()['input_shape']

    @pytest.mark.parametrize(
        "test_layer, input_shape",
        [
            (layers.Dense(16, weight_l2=0.2), (64, 100)),
            (layers.Dropout(0.5), nhwc_shape),
            (layers.Conv2D(32, 3), nhwc_shape),
            (layers.MaxPooling2D(2), nhwc_shape),
            (layers.AvgPooling2D(2), nhwc_shape),
            (layers.BatchNormalization([2, 3]), nhwc_shape),
            (layers.Flatten(), nhwc_shape),
            (activations.ReLU(), (64, 129)),
            (activations.LeakyReLU(), (64, 129)),
            (activations.Sigmoid(), nhwc_shape),
            (activations.Softmax(), (64, 100)),
            (activations.Tanh(), nhwc_shape),
        ]
    )
    def test_from_config(self, test_layer, input_shape):
        new_layer = test_layer.from_config(test_layer.get_config())
        assert id(new_layer) != id(test_layer)
        assert new_layer.get_config() == test_layer.get_config()
        test_layer(ops.ones(input_shape))
        test_layer.training = False
        new_layer = test_layer.from_config(test_layer.get_config(), True)
        assert new_layer.training is False
        assert new_layer.built is True or new_layer.built is None
        assert new_layer.get_config() == test_layer.get_config()
        with pytest.raises(TypeError):
            layers.Layer.from_config(test_layer.get_config())


class TestDense:

    def test_build(self, dense):
        assert dense.weights.shape == (100, 16)
        # Test auto build.
        d = layers.Dense(16, 100, weight_initializer='he')
        assert d.weights.shape == (100, 16)
        # Test build by constructor arguments
        d = layers.Dense(16, 100)
        d.build()
        assert d.weights.shape == (100, 16)
        # Test build when both input_dim and input_shape are passed
        d = layers.Dense(16, 100)
        d.build((64, 128))
        assert d.weights.shape == (128, 16)
        with pytest.raises(ValueError):
            d = layers.Dense(16)
            d.build()

    def test_units(self, dense):
        assert dense.units == 16

    def test_compute_output_shape(self, dense):
        assert dense.compute_output_shape((128, 100)) == (128, 16)
        with pytest.raises(ValueError):
            dense.compute_output_shape(100)

    def test_forward(self, dense):
        i = ops.random.random((128, 100))
        assert dense.forward(i).shape == (128, 16)

    def test_backward(self, dense):
        i = ops.random.random((128, 100))
        dense.inputs = i
        assert dense.backward(dense.forward(i)).shape == i.shape

    def test_get_config(self, dense):
        layer_config = dense.get_config()
        assert layer_config['type'] == 'Dense'
        assert layer_config['neurons'] == 16
        assert len(layer_config) == 9


class TestDropout:

    def test_forward(self):
        layer = layers.Dropout(0.2)
        inputs = ops.ones(nhwc_shape)
        first_pass = layer.forward(inputs)
        assert not ops.array_equal(first_pass, inputs)
        layer.training = False
        assert layer.forward(inputs) is inputs
        layer = layers.Dropout(0)
        assert layer.forward(inputs) is inputs

    def test_backward(self):
        layer = layers.Dropout(0.2)
        inputs = ops.ones(nhwc_shape)
        output = layer.forward(inputs)
        d_inputs = layer.backward(inputs)
        assert ops.array_equal(output, d_inputs)

    def test_get_config(self):
        layer_config = layers.Dropout(0.25).get_config()
        assert layer_config['rate'] == 0.25
        assert len(layer_config) == 4


class TestSpatialLayer:

    # The padding tests aren't for testing if padding behaviour is correct, that is tested for in `test_layer_utils.py`,
    # it's just to make sure it's working as intended from within the layers.
    def test_calculate_padding_amount(self, spatial_layer):
        config.set_image_data_format('channels-last')
        assert sum(spatial_layer(2, 2, 'valid').calculate_padding_amount((0,))) == 0
        assert sum(spatial_layer(2, 2, 'same').calculate_padding_amount(nhwc_shape)) == 0
        assert sum(spatial_layer().calculate_padding_amount(nhwc_shape)) == 2
        config.set_image_data_format('channels-first')
        assert sum(spatial_layer().calculate_padding_amount(nchw_shape)) == 2
        config.set_image_data_format('channels-last')

    def test_padding_amount(self, spatial_layer):
        layer = spatial_layer(3, 2, 'valid')
        layer.input_shape = nhwc_shape
        assert sum(layer.padding_amount) == 0
        layer = spatial_layer()
        layer.input_shape = nhwc_shape
        assert sum(layer.padding_amount) == 2

    def test_nhwc(self, spatial_layer):
        config.set_image_data_format('channels-first')
        assert spatial_layer().nhwc is False
        config.set_image_data_format('channels-last')
        assert spatial_layer().nhwc is True

    def test_compute_output_shape(self, spatial_layer):
        assert spatial_layer().compute_output_shape(nhwc_shape) == (128, 16, 16, 3)
        config.set_image_data_format('channels-first')
        assert spatial_layer().compute_output_shape(nchw_shape) == (128, 3, 16, 16)
        layer = layers.Conv2D(8, 3, 2, 'same')  # Test when channels change.
        assert layer.compute_output_shape(nchw_shape) == (128, 8, 16, 16)
        config.set_image_data_format('channels-last')
        assert layer.compute_output_shape(nhwc_shape) == (128, 16, 16, 8)

    def test_to_nhwc_format(self, spatial_layer):
        assert spatial_layer().to_nhwc_format(nhwc_shape) == nhwc_shape
        config.set_image_data_format('channels-first')
        assert spatial_layer().to_nhwc_format(nchw_shape) == nhwc_shape

    def test_make_arguments_list(self, spatial_layer):
        layer = spatial_layer()
        layer.input_shape = nhwc_shape
        assert len(layer.make_arguments_list(ops.ones(1))) == 13
        pool = layers.MaxPooling2D(2, 2)
        pool.input_shape = nhwc_shape
        assert len(pool.make_arguments_list(ops.ones(1))) == 12


class TestConv2D:

    @pytest.mark.parametrize(
        "input_shape, image_format",
        [
            (nhwc_shape, 'channels-last'),
            (nchw_shape, 'channels-first'),
            (nhwc_shape[1:], 'channels-last'),
            (nchw_shape[1:], 'channels-first')
        ]
    )
    def test_build(self, input_shape, image_format):
        # Test wrong input_shape
        with pytest.raises(ValueError):
            layers.Conv2D(8, 5).build((64, 100))
        config.IMAGE_DATA_FORMAT = image_format
        conv = layers.Conv2D(8, 5)
        conv.build(input_shape)
        assert conv.weights.shape == (8, 5, 5, 3)
        # Test calling build on a built layer.
        with pytest.raises(ValueError):
            conv.build(input_shape)

    @pytest.mark.parametrize(
        "input_shape, dtype, image_format",
        [
            (nhwc_shape, 'float32', 'channels-last'),
            (nhwc_shape, 'd', 'channels-last'),
            (nchw_shape, 'f', 'channels-first'),
            (nchw_shape, 'f4', 'channels-first'),
        ]
    )
    def test_forward(self, input_shape, dtype, image_format):
        config.set_image_data_format(image_format)
        conv = layers.Conv2D(8, 5, 2, 'same')
        conv.build(input_shape)
        config.set_default_dtype(dtype)
        inputs = ops.ones(input_shape)
        output = conv(inputs)
        output_shape = (128, 16, 16, 8) if conv.nhwc else (128, 8, 16, 16)
        assert output.dtype == config.DTYPE
        assert output.shape == output_shape

    @pytest.mark.parametrize(
        "input_shape, dtype, image_format",
        [
            (nhwc_shape, 'float32', 'channels-last'),
            (nhwc_shape, 'f', 'channels-last'),
            (nchw_shape, 'f8', 'channels-first'),
            (nhwc_shape, 'f', 'channels-last'),
            (nhwc_shape, 'd', 'channels-last'),
            (nchw_shape, '<f4', 'channels-first'),
        ]
    )
    def test_backward(self, input_shape, dtype, image_format):
        config.set_image_data_format(image_format)
        conv = layers.Conv2D(8, 5, 2, 'same', 0.02, 0.02)
        ops.random.seed(0)  # Set seed to 0 so the random results are consistent for testing purposes.
        config.set_default_dtype(dtype)
        conv.build(input_shape)
        conv.input_shape = input_shape
        conv.inputs = ops.ones(conv.padded_input_shape)
        d_values = ops.ones(conv.output_shape)
        d_w_r_t_inputs = conv.backward(d_values)  # Derivative w.r.t inputs.
        assert d_w_r_t_inputs.shape == input_shape
        assert d_w_r_t_inputs.dtype == config.DTYPE
        # That numbers that are checked against have been obtained from calculating the gradients w.r.t to the input
        # passed to a TensorFlow Conv2D layer constructed with the same parameters and has the same weights as the
        # Conv2D layer above, so if TensorFlow's numbers are correct (I'm assuming they are cuz duh) and the results of
        # `d_w_r_t_inputs` are close to them, means they are correct too.
        assert d_w_r_t_inputs.min() == pytest.approx(-3.5361599922180176)
        assert d_w_r_t_inputs.max() == pytest.approx(6.143856048583984)
        assert d_w_r_t_inputs.mean() == pytest.approx(1.849550110947651)

    @pytest.mark.parametrize(
        'input_shape, image_data_format',
        [
            (nhwc_shape, 'channels-last'),
            (nchw_shape, 'channels-first')
        ]
    )
    def test_set_params(self, input_shape, image_data_format):
        config.set_image_data_format(image_data_format)
        conv = layers.Conv2D(8, 5, 2, 'same', 0.02)
        conv.build(input_shape)
        weights, biases = conv.get_params()
        # Test shape mismatch.
        with pytest.raises(ValueError):
            conv.set_params([weights.transpose((1, 2, 3, 0))])
        # Test correct handling for keras Conv2D weights.
        conv.set_params([weights.transpose((1, 2, 3, 0))], True)
        assert conv.weights.shape == (8, 5, 5, 3)
        conv.set_params([weights.transpose((1, 2, 3, 0)), biases], True)
        assert conv.weights.shape == (8, 5, 5, 3)

    def test_get_config(self):
        layer_config = layers.Conv2D(16, 3, 2, 'same', 0.2).get_config()
        assert len(layer_config) == 11
        assert layer_config['type'] == 'Conv2D'
        assert layer_config['kernels'] == 16
        assert layer_config['kernel_size'] == (3, 3)


class TestPooling2D:

    @pytest.mark.parametrize(
        "pool_size, strides, padding, expected_shape",
        [
            (2, 3, 'same', (128, 11, 11, 3)),
            (2, 1, 'valid', (128, 31, 31, 3)),
            (2, 2, 'valid', (128, 16, 16, 3)),
            (2, 1, 'same', (128, 32, 32, 3)),
            (1, 1, 'valid', (128, 32, 32, 3)),
            (1, 2, 'valid', (128, 16, 16, 3))
        ]
    )
    def test_forward(self, pool_size, strides, padding, expected_shape):
        for layer_type in (layers.MaxPooling2D, layers.AvgPooling2D):
            for input_shape, image_format in zip((nhwc_shape, nchw_shape), ('channels-last', 'channels-first')):
                config.set_image_data_format(image_format)
                dtype = ops.random.choice(('f', 'd'))
                config.set_default_dtype(dtype)
                layer = layer_type(pool_size, strides, padding)
                output = layer(ops.ones(input_shape))
                assert output.dtype == dtype
                if image_format == 'channels-first':
                    assert output.shape == (expected_shape[0], expected_shape[3], *expected_shape[1:3])
                else:
                    assert output.shape == expected_shape
                if sum(layer.padding_amount):
                    # If the type of the layer is average pooling and the inputs were padded, the mean is going to be
                    # less than one because the inputs were padded with zeros.
                    if layer_type == layers.AvgPooling2D:
                        assert output.mean() < 1.
                        break
                assert output.mean() == 1.

    @pytest.mark.parametrize(
        "pool_size, strides, padding",
        [
            (3, 4, 'same'),
            (2, 1, 'same'),
            (2, 2, 'valid'),
            (1, 1, 'valid'),
            (1, 2, 'valid'),
            (2, 1, 'same')
        ]
    )
    def test_backward(self, pool_size, strides, padding):
        for layer_type in (layers.MaxPooling2D, layers.AvgPooling2D):
            for input_shape, image_format in zip((nhwc_shape, nchw_shape), ('channels-last', 'channels-first')):
                config.set_image_data_format(image_format)
                dtype = ops.random.choice(('f', 'd'))
                config.set_default_dtype(dtype)
                layer = layer_type(pool_size, strides, padding)
                layer.input_shape = input_shape
                if layer_type == layers.MaxPooling2D:  # We need to perform a forward pass in max pooling case to
                    # compute the masks (array of locations of where the max values were).
                    d_values = layer(ops.ones(input_shape))
                else:
                    d_values = ops.ones(layer.output_shape)
                output = layer.backward(d_values)
                assert output.dtype == dtype
                assert output.shape == input_shape

    def test_get_params(self):
        with pytest.raises(TypeError):
            layers.MaxPooling2D(3).get_params()

    def test_set_params(self):
        with pytest.raises(TypeError):
            layers.MaxPooling2D(3).set_params([])

    def test_get_config(self):
        max_config = layers.MaxPooling2D(2).get_config()
        avg_config = layers.AvgPooling2D(2).get_config()
        assert len(max_config) == len(avg_config) == 6
        assert max_config['type'] == 'MaxPooling2D'
        assert avg_config['type'] == 'AvgPooling2D'
        assert max_config['pool_size'] == (2, 2)
        assert max_config['strides'] == (1, 1)


class TestBatchNormalization:

    @pytest.mark.parametrize(
        # "axis, n_dims, image_format, expected",
        "args",
        [
            (None, 4, 'channels-last', (0, 1, 2)),
            (None, 4, 'channels-first', (0, 2, 3)),
            (1, 4, 'channels-last', (0, 2, 3)),
            ((1, 2, 3), 4, 'channels-last', (0,)),
            ((0, 1, 2), 4, 'channels-last', (3,)),
            ((0, 1, 2), 4, 'channels-first', (3,)),
            (None, 2, 'channels-first', (0,)),
            ((1, 2), 4, 'channels-first', (0, 3)),
            ((1, 2), 2, 'channels-first', (0, 3), pytest.raises(ValueError)),
            ((1, 2), 3, 'channels-last', (0, 3), pytest.raises(ValueError)),
            ((-1, -2), 4, 'channels-last', (0, 1)),
            ((-4), 4, 'channels-last', (1, 2, 3)),
            ((1, 0), 4, 'channels-last', (2, 3)),
            ([3], 4, 'channels-last', (0, 1, 2)),
            ((3,), 4, 'channels-last', (0, 1, 2))
        ]
    )
    def test_get_reduction_axis(self, args):
        config.set_image_data_format(args[2])
        cnxt_mngr = nullcontext() if len(args) == 4 else args[-1]
        with cnxt_mngr:
            assert layers.BatchNormalization(args[0]).get_reduction_axis(args[1]) == args[3]

    @pytest.mark.parametrize(
        "input_shape, image_format, expected_shape",
        [
            (nhwc_shape, 'channels-last', (1, 1, 1, 3)),
            (nhwc_shape[1:], 'channels-last', (1, 1, 1, 3)),
            (nchw_shape, 'channels-first', (1, 3, 1, 1)),
            (nchw_shape[1:], 'channels-first', (1, 3, 1, 1)),
            ((128, 100), 'channels-first', (1, 100)),
            ((128, 64), 'channels-last', (1, 64)),
            (64, 'channels-last', (1, 64))
        ]
    )
    def test_build(self, input_shape, image_format, expected_shape):
        config.set_image_data_format(image_format)
        batch_norm = layers.BatchNormalization()
        batch_norm.build(input_shape)
        assert batch_norm.weights.shape == batch_norm.biases.shape == expected_shape

    @pytest.mark.parametrize(
        "axis, input_shape, image_format, dtype",
        [
            (None, nhwc_shape, 'channels-last', 'f'),
            (None, nchw_shape, 'channels-first', 'f'),
            (None, nchw_shape, 'channels-first', 'd'),
            ((1, 3), nchw_shape, 'channels-first', 'f'),
            ((1, 2, 0), nchw_shape, 'channels-first', 'f'),
            ((1, 2, 0), nhwc_shape, 'channels-last', 'f'),
            (None, (64, 100), 'channels-first', float),
            (0, (64, 100), 'channels-first', 'float32'),
        ]
    )
    def test_forward(self, axis, input_shape, image_format, dtype):
        config.set_image_data_format(image_format)
        config.set_default_dtype(dtype)
        output = layers.BatchNormalization(axis)(ops.random.uniform(-10, 16, input_shape))
        # pytest absolute tolerance is 1e-12, this value is too small for float32 arrays but is fine for float64 arrays.
        tolerance = 5e-5 if config.DTYPE == 'float32' else None
        assert output.mean() == pytest.approx(0, abs=tolerance)
        assert output.std() == pytest.approx(1, abs=tolerance)

    def test_get_params(self):
        config.set_image_data_format('channels-last')
        layer = layers.BatchNormalization()
        # Test before build
        with pytest.raises(ValueError):
            layer.get_params()
        layer.build(nhwc_shape)
        # Test that we got the parameters correctly.
        assert all([param.shape == (1, 1, 1, nhwc_shape[-1]) for param in layer.get_params()])
        # Test that weights are passed by reference when copy is False, meaning they can be changed by the user.
        weights = layer.get_params(False)[0]
        weights += 1
        assert ops.array_equal(layer.weights, weights)
        # Test that copy actually copies the arrays.
        weights = layer.get_params(True)[0]
        weights += 1
        assert not ops.array_equal(layer.weights, weights)
        # Test that squeeze works.
        assert all([param.shape == (nhwc_shape[-1],) for param in layer.get_params(True, True)])
        layer = layers.BatchNormalization((1, 2))
        layer.build(nhwc_shape)
        assert all([param.shape == (1, *nhwc_shape[1:3], 1) for param in layer.get_params(True, True)])

    @pytest.mark.parametrize(
        'input_shape, image_data_format',
        [
            (nhwc_shape, 'channels-last'),
            (nchw_shape, 'channels-first')
        ]
    )
    def test_set_params(self, input_shape, image_data_format):
        config.set_image_data_format(image_data_format)
        layer = layers.BatchNormalization()
        # Test before the layer is built.
        with pytest.raises(ValueError):
            layer.set_params([])
        layer.build(input_shape)
        before_params = layer.get_params()
        layer.set_params([ops.ones(param.shape) for param in before_params])
        # Test that all the layer's parameters actually changed.
        assert not all([ops.array_equal(
            before_param, after_param) for before_param, after_param in zip(before_params, layer.get_params())])
        # Test setting only one element.
        before_weights = layer.weights
        layer.set_params([ops.random.standard_normal(layer.weights.shape)])
        assert not ops.array_equal(layer.weights, before_weights)
        # Test wrong shape.
        with pytest.raises(ValueError):
            layer.set_params([ops.ones(layer.weights.shape), ops.ones(4)])
        with pytest.raises(ValueError):
            layer.set_params([ops.ones((1, 1, 1, 4))])
        # Test one dimensional parameters that match the layer's parameter shapes.
        layer.set_params(layer.get_params(squeeze=True))
        expected_shape = (1, 3, 1, 1) if image_data_format == 'channels-first' else (1, 1, 1, 3)
        assert layer.weights.shape == expected_shape

    def test_get_config(self):
        layer_config = layers.BatchNormalization().get_config()
        assert len(layer_config) == 9
        assert layer_config['type'] == 'BatchNormalization'



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
from xrnn import activations
from xrnn import ops
import pytest

inputs = ops.random.uniform(-2, 2, (10, 10))


class TestReLU:

    @pytest.mark.parametrize("alpha", [-0.01, 1])
    def test_alpha_arg(self, alpha: float) -> None:
        with pytest.raises(ValueError):
            activations.ReLU(alpha)

    def test_forward(self):
        relu = activations.ReLU()
        output = relu.forward(inputs)
        assert ops.all(ops.argwhere(output == 0) == ops.argwhere(inputs <= 0))

    def test_backward(self):
        relu = activations.ReLU()
        relu.forward(inputs)
        d_output = relu.backward(inputs)
        assert ops.all(ops.argwhere(d_output > 0) == ops.argwhere(inputs > 0))


class TestLeakyReLU:

    def test_forward(self):
        lrelu = activations.LeakyReLU()
        output = lrelu.forward(inputs)
        assert ops.all(ops.argwhere(output == (inputs * 0.01)) == ops.argwhere(inputs <= 0))

    def test_backward(self):
        lrelu = activations.LeakyReLU()
        lrelu.forward(inputs)
        d_output = lrelu.backward(inputs)
        assert ops.allclose(d_output, inputs * ops.where(inputs > 0, 1, 0.01))


class TestSoftmax:

    @pytest.mark.parametrize(
        "input_batch, expect",
        [
            (inputs, nullcontext()),  # Softmax expects inputs to be 2-dimensional, so this one should pass.
            (ops.random.uniform(0, 1, (10, 10, 10)), pytest.raises(ValueError))  # But this shouldn't
        ]
    )
    def test_input_dims(self, input_batch, expect):
        with expect:
            assert activations.Softmax()(input_batch) is not None



"""This module defines variables and functions that can alter the behavior of the whole package like setting/changing
the default data type used to store arrays and perform calculations, and changing how layers treat images by calling
`set_image_data_format(data_format)`."""
import ctypes
import os
import sys
import warnings
from typing import Union, Set

import numpy as np
import numpy.__config__ as np_config

if sys.version_info.minor < 8:
    from typing_extensions import Literal
else:
    from typing import Literal  # Literal was added to typing in python 3.8.

DtypeHint = Union[Literal['float32', 'f', 'f4', '<f4', 'single', 'float64', 'f8', '<f8', 'double', 'float', 'd'], type]
# The default values to use across the whole package.
EPSILON: float = 1e-7  # A good value for float32, and it's changed to 1e-14 for float64.
IMAGE_DATA_FORMAT: Literal['channels-last', 'channels-first'] = 'channels-last'
DTYPE: DtypeHint = 'float32'
CREATED_OBJECTS: list = []  # A list that contains all objects that should be kept track of that have been created.
SEEN_NAMES: Set[str] = set()  # A list tracking the names that have been seen during the session.
MKL_WARNING_MSG: str = (
    "KMP_DUPLICATE_LIB_OK environment variable has been set to True because numpy is using Intel's MKL library as a "
    "BLAS backend. This is needed for this module to work because it uses a different parallelization implementation "
    "to Intel's which causes `may` cause instabilities or crashes. For more information refer to "
    "`handle_mkl_numpy_openmp_conflict` function documentation in `config.py`")
# This is needed because resource warnings are ignored (not shown) by default. This filter is used to only
# show this specific ResourceWarning once and keep the default behaviour, which is `ignore`, unchanged.
warnings.filterwarnings('once', MKL_WARNING_MSG, ResourceWarning)


def parse_datatype(dtype: DtypeHint) -> Literal['float32', 'float64']:
    """Parses the datatype from a multitude of sources and return either 'float32' or 'float64'. This function can take
    a lot of different sources as the `dtype` argument such as numpy arrays, numpy `dtype` objects, `.dtype` property
    of numpy objects that have it, string literals, python `float`, numpy datatype objects like `np.float32` or
    `np.single`, etc. If you happen to pass a thing that can't be parsed the function will be pretty verbose."""
    if dtype is float:
        return 'float64'
    if dtype is np.float32 or dtype is np.single:
        return 'float32'
    if dtype is np.float64 or dtype is np.float_ or dtype is np.double:
        return 'float64'
    if not isinstance(dtype, str):
        if hasattr(dtype, 'dtype'):
            dtype = str(dtype.dtype)
        else:
            dtype = str(dtype)
    if dtype in ('float32', 'f', 'f4', '<f4', 'single'):
        return 'float32'
    if dtype in ('float64', 'f8', '<f8', 'double', 'float', 'd'):
        return 'float64'
    if '>' in dtype:
        raise TypeError("Big endian byte order not supported.")
    raise TypeError(
        f"Expected one of 'float32', 'f', 'f4', '<f4', 'single', 'float64', 'f8', '<f8', 'double', 'float', "
        f"but got {dtype}. For a description of acceptable input arguments, look at the function documentation.")


def set_default_dtype(dtype: DtypeHint = 'float32') -> None:
    """Set a new default data type. Changing the data type will change the dtype for all subsequent operations after
    the function call. This function should only be used once preferably before using the model (train/inference)
    because it might lead to unstable training due to rounding errors and precision loss when converting data types."""
    global DTYPE, EPSILON
    dtype = parse_datatype(dtype)
    if dtype == DTYPE:
        return
    DTYPE = dtype
    EPSILON = 1e-14 if DTYPE == 'float64' else 1e-7
    for trackable in CREATED_OBJECTS:
        trackable.dtype = DTYPE


def set_image_data_format(
        image_format: Literal['channels-first', 'channels-last', 'channels_first', 'channels_last']) -> None:
    """Sets how the images should be treated. If 'channels-last' (default), images are expected to have a shape
    of (batch_size, height, width, channels) or (NHWC), if 'channels-first' the expected shape is (NCHW).
    *Note* that the execution speed and model performance (loss) might differ slightly between the two image formats."""
    # The function takes the underscore form of the inputs to stay consistent with tensorflow/keras way of doing this.
    # However, it changes the underscore to a hyphen because that's how they are interpreted internally.
    if image_format not in ('channels-first', 'channels-last', 'channels_first', 'channels_last'):
        raise ValueError(
            f"`image_format` must one of 'channels-first', 'channels-last', 'channels_first', 'channels_last'. "
            f"Got {image_format} instead.")
    image_format = image_format.replace("_", "-")
    global IMAGE_DATA_FORMAT
    IMAGE_DATA_FORMAT = image_format


def set_epsilon(value: float) -> None:
    """Sets the epsilon to `value` that will be used throughout the whole codebase. **Make** sure that epsilon value is
    small enough to not affect the results of calculations and not too small that it can't prevent division by zero."""
    global EPSILON
    EPSILON = value


def set_sleep_state(allow_sleep: bool = True) -> None:
    """Prevents the PC from sleeping during the model training. *Note* only works on Windows."""
    prevent_sleep_flag = 0x00000001  # Prevents the pc from sleeping but allows the screen to turn off.
    reset_flag = 0x80000000  # Resets the sleep status to what they were before (Windows remembers the initial state).
    flag = reset_flag if allow_sleep else prevent_sleep_flag | reset_flag
    if os.name == 'nt':
        ctypes.windll.kernel32.SetThreadExecutionState(flag)


def handle_mkl_numpy_openmp_conflict() -> None:
    """
    Sets the KMP_DUPLICATE_LIB_OK environment variable to True, this is required in case that the installed numpy is
    using Intel's MKL as its backend for accelerating math operations, usually happens when numpy is installed via
    conda, mainly `np.dot`. The reason this is required is that MKL uses Intel's own OpenMP library, which conflicts
    with the OpenMP implementation used in this package (MSVC or MinGW) causing OpenMP to raise an error that multiple
    copies of it are linked to the same program. This environment variable is a `workaround` to allow the program to
    continue to execute.

    .. note::
       This can be dangerous, according to the error raised by OpenMP which states the following: "That is dangerous,
       since it can degrade performance or cause incorrect results ... [setting KMP_DUPLICATE_LIB_OK=True] may cause
       crashes or silently produce incorrect results".
       But from my testing, it hasn't caused any of the aforementioned possible complications.
    """
    mkl_numpy_defined_variable_names = ('blas_mkl_info', 'lapack_mkl_info')
    for name in mkl_numpy_defined_variable_names:
        if getattr(np_config, name, None):  # Checks if numpy is using MKL.
            os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
            warnings.warn(MKL_WARNING_MSG, ResourceWarning)
            return


handle_mkl_numpy_openmp_conflict()



"""This module contains some utility functions that are used by all sorts of layers."""
from typing import Union, List

from xrnn import config
from xrnn import ops


def compute_spatial_output_shape(
        input_shape: tuple, window_size: tuple, strides: tuple, padding_amount: tuple) -> tuple:
    """
    Computes the resulting output height and width from convolution and pooling layers after performing their
    calculations on the input.

    Parameters
    ----------
    input_shape: tuple
        (batch_size, height, width, channels) if images are in NHWC format, (batch_size, channels, height, width) if
        images are in NCHW format.
    window_size: tuple
        (kernel or pool window height, kernel or pool window width).
    strides: tuple
        (strides height, strides width).
    padding_amount: tuple
        The amount of padding on each side (pad_top, pad_bot, pad_left, pad_right).

    Returns
    -------
    new_spatial_output: tuple
        (new height, new width).
    """
    if len(padding_amount) != 4:
        raise ValueError(
            f'`padding_amount` must be a tuple containing four elements: pad_top, pad_bot, pad_left, pad_right. '
            f'Got {padding_amount}.')
    if config.IMAGE_DATA_FORMAT == 'channels-last':
        input_shape = (input_shape[1], input_shape[2])
    else:
        input_shape = input_shape[2:]
    return tuple(
        (input_shape[i] + sum(padding_amount[i * 2:i * 2 + 2]) - window_size[i]
         ) // strides[i] + 1 for i in range(len(input_shape)))


def calculate_padding_on_sides(input_shape: tuple, window_size: tuple, strides: tuple) -> tuple:
    """
    Calculates the padding value for each side of the image and returns (padding_top, padding_bottom,
    padding_left, padding_right).

    Parameters
    ----------
    input_shape: tuple
        (batch_size, height, width, channels) if images are in NHWC format, (batch_size, channels, height, width) if
        images are in NCHW format.
    window_size: tuple
        (kernel or pool window height, kernel or pool window width).
    strides: tuple
        (strides height, strides width).

    Returns
    -------
    padding_amount: tuple of four ints
        (padding_top, padding_bottom, padding_left, padding_right).

    Notes
    -----
    This is implemented the same way that `tensorflow` calculates padding, and it's different from other deep learning
    libraries like `cuDNN` and `Caffe` [1]_.

    References
    ----------
    .. [1] `https://mmuratarat.github.io/2019-01-17/implementing-padding-schemes-of-tensorflow-in-python`.
    """
    padding_size = []
    if config.IMAGE_DATA_FORMAT == 'channels-last':
        input_shape = (input_shape[1], input_shape[2])
    else:
        input_shape = input_shape[2:]
    for i in range(2):
        reminder = input_shape[i] % strides[i]
        if reminder == 0:
            padding = window_size[i] - strides[i]
        else:
            padding = window_size[i] - reminder
        padding = max(padding, 0)
        padding_size.append((padding // 2, padding - (padding // 2)))
    return tuple(sum(padding_size, ()))  # Unpack the tuples.


def to_tuple(maybe_tup: Union[int, tuple]) -> tuple:
    """Converts an integer to a tuple of (integer, integer) if not already a tuple. Used for integer kernel/pool size
    and strides."""
    if isinstance(maybe_tup, int):
        return maybe_tup, maybe_tup
    return tuple(maybe_tup)


def validate_padding(padding: config.Literal['same', 'valid']) -> config.Literal['same', 'valid']:
    """Makes sure that padding mode is a valid one, meaning it's one of 'same' or 'valid', and returns it."""
    if padding not in ('same', 'valid'):
        raise ValueError(f"`padding` must be 'same' or 'valid'. Got {padding} instead.")
    return padding


def padded_input_shape(input_shape: tuple, padding_amount: tuple) -> tuple:
    """
    Returns the padded input shape

    Parameters
    ----------
    input_shape: tuple
        input shape to be padded.
    padding_amount: tuple
        A tuple of four ints specifying the padding amount on each side of the input.

    Returns
    -------
    padded_input_shape: tuple
        The shape of the padded input.
    """
    ph = sum(padding_amount[:2])
    pw = sum(padding_amount[2:])
    if config.IMAGE_DATA_FORMAT == 'channels-last':
        return input_shape[0], input_shape[1] + ph, input_shape[2] + pw, input_shape[3]
    return input_shape[0], input_shape[1], input_shape[2] + ph, input_shape[3] + pw


def pad_batch(inputs: ops.ndarray, padding_dims: tuple) -> ops.ndarray:
    """Zero pads the inputs from all sides (top, bottom, left, right) along the height and width axis."""
    if not sum(padding_dims):
        return inputs
    padding_top, padding_bottom, padding_left, padding_right = padding_dims
    if config.IMAGE_DATA_FORMAT == 'channels-last':
        padding = ((0, 0), (padding_top, padding_bottom), (padding_left, padding_right), (0, 0))
    else:
        padding = ((0, 0), (0, 0), (padding_top, padding_bottom), (padding_left, padding_right))
    # wrapped the tuple with padding values for each side with a numpy array so pycharm type checker can stop crying,
    # and it was going to be turned into numpy array by numpy anyway, so its impact on performance should be negligible.
    return ops.pad(inputs, ops.array(padding))


def extract_from_padded(inputs: ops.ndarray, padding_dims: tuple) -> ops.ndarray:
    """Returns the original array from the padded array."""
    padding_top, padding_bottom, padding_left, padding_right = padding_dims
    # If either padding_bottom or padding_right are zero that means there's no padding along the axis they
    # correspond to because they are always the larger number and if they are equal to zero then the others are too.
    if padding_bottom:
        inputs = inputs[:, padding_top:-padding_bottom] \
            if config.IMAGE_DATA_FORMAT == 'channels-last' else inputs[:, :, padding_top:-padding_bottom]
    if padding_right:
        inputs = inputs[:, :, padding_left:-padding_right] \
            if config.IMAGE_DATA_FORMAT == 'channels-last' else inputs[..., padding_left:-padding_right]
    if not inputs.flags['C']:
        inputs = ops.ascontiguousarray(inputs, config.DTYPE)
    return inputs


def layer_memory_consumption(layer, input_shape: tuple = None, training: bool = True, adam: bool = False) -> tuple:
    """
    Calculates a layer's memory consumption separately for each of the following:
     1. Parameters: Memory consumed by the layer's weights and biases if it has them, if not it's zero.
     2. Gradients: Memory consumed by the layer's gradients, which are the derived weights, biases and some values
        created by the optimizer. *Note* that Adam uses slightly more memory than other optimizers.
     3. Activations: Activations memory consumption doesn't calculate what it implies on first sight, which is memory
        consumption for activation function layers. It actually calculates the memory consumed by the layer inputs' and
        any intermediate results during the forward pass because the layer will use them again during the backward pass.
        Activation memory consumption is zero if the layer is in inference mood (when training is set to False) since
        the inputs/intermediates aren't going to be saved because there's no backward pass.
     4. Total memory consumption: The sum of the aforementioned.

    Parameters
    ----------
    layer: Layer subclass or instance
        The layer to calculate its memory consumption.
    input_shape: tuple, optional
        Input shape to the layer, this is needed to calculate the saved activation memory consumption. The first axis
        should be the `batch_size`. If not provided, the `input_shape` attribute of the layer will be used.
    training: bool, optional
        Whether to calculate gradient and activation memory consumption, because they are only needed when training.
    adam: bool, optional
        Whether adam optimizer is being used for training, this needs to be explicitly provided because 'adam' consumes
        slightly more memory than other optimizers.

    Returns
    -------
    mem_consumption: tuple of four floats.
        parameters (weights + biases), gradients, saved activations, total memory consumption in bytes.
    """
    parameters_mem = layer.weights.nbytes + layer.biases.nbytes if hasattr(layer, 'weights') else 0
    gradients_mem, intermediates_mem = 0, 0
    # Stores memory consumption for the derived weights and biases and arrays created by the optimizer to calculate them
    if training:
        gradients_mem = parameters_mem * 3 if adam else parameters_mem * 2
        if getattr(layer, 'padding', None) == 'same':
            input_shape = padded_input_shape(
                input_shape, calculate_padding_on_sides(input_shape, layer.window_size, layer.strides))
        intermediates_mem = ops.prod(input_shape) * ops.dtype(config.DTYPE).itemsize
        intermediates_mem *= 2 if 'Dropout' in layer.name or 'BatchNormalization' in layer.name else 1
    parameters_mem += layer.weights.nbytes * 2 if 'BatchNormalizatio' in layer.name and hasattr(layer, 'weights') else 0
    return parameters_mem, gradients_mem, intermediates_mem, parameters_mem + gradients_mem + intermediates_mem


def make_unique_name(obj: object) -> str:
    """Makes a given object name unique during a session."""
    if isinstance(obj, type):
        raise TypeError("The passed object must be initialized.")
    name = type(obj).__name__
    identifier = 0
    for seen_name in config.SEEN_NAMES:
        if seen_name.split('_')[0] == name.split('_')[0]:
            identifier += 1
    name += f"_{identifier}"
    config.SEEN_NAMES.add(name)
    return name


def to_readable_unit_converter(num: Union[int, float], n_digits: int = 2) -> str:
    """
    Prints the number of bytes in a human-readable format.

    .. note::
       This method prefixes the units with 'iB', because it calculates the units using powers of 2 (IEC) not 10 (SI),
       that's why it prints `Kib` and not `KB` and I don't want to add the confusion, so I'm sticking with the standard
       for the calculations and unit names, please don't argue with me, it's not my fault.
    """
    # Thanks to https://stackoverflow.com/a/1094933
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.{n_digits}f} {unit}B"
        num /= 1024.0
    return f"{num:.1f} YiB"


def print_table(table: List[List], additional_notes: dict) -> None:
    """Prints a nicely formatted table from a list of lists which represents rows and columns."""
    padding = [0] * 4
    for row in table:
        for i, value in enumerate(row):
            padding[i] = max(len(str(value)), padding[i])
    line_length = sum(padding) + 4 * 4  # 4 (value per row) * 4 (spaces between columns)
    print('-' * line_length)
    for i, row in enumerate(table):
        for j, value in enumerate(row):
            print(f"{value:<{padding[j]}}", end=' ' * 4)
        if i == 0:
            print('\n' + '=' * line_length)
        else:
            print('\n')
    print('=' * line_length)
    for key, value in additional_notes.items():
        print(f"{key}: {value}")
    print('-' * line_length)


def time_unit_converter(time: float) -> str:
    """Gets the time in seconds and returns a string representing the time in a nice ETA way. For e.g. 180.30 is
    converted to 3.0 minutes."""
    if time < 1:
        return f"{time * 100:.0f} ms"
    if 1 <= time < 60:
        return f"{time:2.0f} sec"
    if 60 <= time < 3600:
        return f"{time / 60:2.1f} min"
    if 3600 <= time < (60 * 60 * 24):
        return f"{time / 60 / 60:2.1f} hrs"
    return f"{time / 60 / 60 / 24:3.2f} days"



"""Defines the `Loss` base class that all loss classes should derive from, `CategoricalCrossentropy` loss class,
`BinaryCrossentropy` loss class and `MSE` or `MeanSquaredError` loss class."""
from xrnn import config
from xrnn import ops


class Loss:

    def __init__(self) -> None:
        """
        The base class that all other loss class should be derived from. Subclasses of this class should define:
         1. `forward` method that takes two numpy arrays, `y_true` and `y_pred`, as input and returns the loss for each
            sample in the batch.
         2. `backward` method that takes two numpy arrays, `y_true` and `d_values (the output of `forward`) and returns
            a numpy array of the gradients w.r.t. each sample in the batch.
        """
        self.trainable_layers = []
        self.accumulate_loss = 0
        self.accumulate_count = 0

    def regularization_loss(self) -> float:
        """Calculates the regularization loss for all the layers that use it."""
        regularization_loss = 0.
        for layer in self.trainable_layers:
            if layer.weight_l2:
                regularization_loss += layer.weight_l2 * ops.sum(ops.square(layer.weights))
            if layer.bias_l2:
                regularization_loss += layer.bias_l2 * ops.sum(ops.square(layer.biases))
        return regularization_loss

    def reset_count(self) -> None:
        """Resets any saved loss values and step count to prepare for a new epoch."""
        self.accumulate_loss = 0
        self.accumulate_count = 0

    def forward(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        """Calculates the loss sample wise using the loss function. Calculating the loss should be done by calling the
        loss class `'loss(y_true, y_pred)'` not by calling `forward` directly."""
        raise NotImplementedError('This method must be overridden.')

    def calculate(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> float:
        """
        A uniform way to calculate the loss and regularization loss as a single number.

        Notes
        -----
        This method calculates an accumulated loss from all batches not only the current batch loss, which results in a
        smoother trend. The backward pass is still performed every batch.
        """
        sample_losses = self(y_true, y_pred)
        self.accumulate_loss += ops.sum(sample_losses)
        self.accumulate_count += len(sample_losses)
        return self.accumulate_loss / self.accumulate_count + self.regularization_loss()

    def backward(self, y_true: ops.ndarray, d_values: ops.ndarray) -> ops.ndarray:
        """Calculates the gradients w.r.t loss and returns the value."""
        raise NotImplementedError("This method must be overridden.")

    def __call__(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        """Calculates the loss sample wise using the `forward` method of the loss function class. This method should be
        called not `forward` because it performs some modifications on the input before passing to `forward`."""
        if not isinstance(self, MSE):
            # Clip data to prevent division by 0.
            # Clip both sides to not drag mean towards any value.
            y_pred = ops.clip(y_pred, config.EPSILON, 1 - config.EPSILON)
        if not isinstance(self, CategoricalCrossentropy):
            if y_true.ndim == 1:
                y_true = ops.expand_dims(y_true, 1)
        return self.forward(y_true, y_pred)


class CategoricalCrossentropy(Loss):
    """
    Categorical Crossentropy loss class. Passed to `model.set(loss=CategoricalCrossentropy())` to use it. Categorical
    Crossentropy loss is used for classification when there are more than two classes in the dataset. For two class
    classification, see `~BinaryCrossentropy`. `y_true` (the labels) can be a one-hot encoded vector or integer classes.
    """

    def forward(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        # CategoricalCrossentropy = -log(yik), k: correct class index
        if y_true.ndim == 2:
            if y_true.shape[1] == 1:  # If the second dimension is 1, we can get rid of it, because it's either integer
                # classes or just two classes, we can safely assume that's the case because for the labels to be one hot
                # encoded, the second dimension should at least be 2.
                y_true = ops.squeeze(y_true.copy())  # Need to copy because squeeze returns the same array/view of it.
            else:  # if labels are one-hot encoded convert them to a sparse array
                y_true = ops.argmax(y_true, axis=1)
        try:
            correct_confidences = y_pred[ops.arange(len(y_true)), y_true.astype('int')]
        except IndexError:  # Try blocks are really cheap when no exceptions are raised. Much cheaper than checking
            # for the maximum value in y_true and comparing that to the length of the 2nd dimension of y_pred.
            raise IndexError(
                "The number of classes in the labels is greater than the number of neurons in the classification "
                "layer. Please decrease the number of classes or add more neurons to the classification layer.")
        negative_log_likelihoods = -ops.log(correct_confidences)
        return negative_log_likelihoods

    def backward(self, y_true: ops.ndarray, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of CategoricalCrossentropy is: -y_true / y_hat.
        # If labels are sparse, turn them into one-hot vector
        if y_true.ndim == 1 or y_true.shape[1] == 1:
            y_true = ops.eye(len(d_values[0]))[y_true.squeeze()]
        d_values = ops.clip(d_values, config.EPSILON, ops.amax(d_values) - config.EPSILON)  # To avoid division by zero.
        return -y_true / d_values / len(d_values)  # The loss derivative. Normalize the gradients.
        # We normalize the gradients, so it’ll become invariant to the number of samples we calculate it for. Necessary
        # when performing the optimization process since the optimizer will sum all the gradients we do this so the
        # sum's magnitude becomes invariant to the number of samples.


class BinaryCrossentropy(Loss):
    """
    Binary Crossentropy loss function class. Used for binary classification, like for classifying if an input image is a
    dog or a cat (or if it's a cat or not cat). Therefor classification layer, which is a `Dense` layer, can have either
    2 neurons (cat or dog) or 1 neuron (cat or not cat).
    """

    def forward(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        # BinaryCrossentropy = -y_true * log(y_pred) + (1 - y_true) * log(1 - y_pred)
        sample_losses = -(y_true * ops.log(y_pred) + (1 - y_true) * ops.log(1 - y_pred))
        sample_losses = ops.mean(sample_losses, axis=-1)
        return sample_losses

    def backward(self, y_true: ops.ndarray, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of BinaryCrossentropy is: -1/N * (y_true/y_hat - (1 - y_true)/(1 - y_hat))
        if y_true.ndim == 1:
            y_true = ops.expand_dims(y_true, 1)
        clipped_dvalues = ops.clip(d_values, config.EPSILON, 1 - config.EPSILON)  # To avoid division by zero.
        return -(y_true / clipped_dvalues - (1 - y_true) / (1 - clipped_dvalues)) / len(d_values[0]) / len(d_values)


class MSE(Loss):
    """Mean Squared Error loss function class. Used for regression problems with continuous output like modeling sin()
    or predicting the temperature."""

    def forward(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        # MSE = 1/N * sum((y_true - y_hat) ** 2)
        return ops.mean(ops.square(y_true - y_pred), axis=-1)  # Sample wise loss.

    def backward(self, y_true: ops.ndarray, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of mse is: -2/N * (y_true - y_hat)
        if y_true.ndim == 1:
            y_true = ops.expand_dims(y_true, 1)
        return -2 * (y_true - d_values) / len(d_values[0]) / len(d_values)


MeanSquaredError = MSE  # Alias



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
from xrnn.data_handler import DataHandler
from xrnn import ops
import pytest


class OnlyLen:
    def __len__(self):
        return 10


class OnlyGetitem:
    def __getitem__(self, idx):
        pass


class ReturnThree(OnlyLen):
    def __getitem__(self, idx):
        return 1, 2, 3


class ReturnOneNonIterable(OnlyLen):
    def __getitem__(self, idx):
        return 2


class ReturnOneIterable(OnlyLen):
    def __getitem__(self, idx):
        return ops.ones(1)


class ReturnTwoWrongType(OnlyLen):
    def __getitem__(self, idx):
        return 1, 2


class ReturnDifferentLength(OnlyLen):
    def __getitem__(self, idx):
        return x[:2], y[:3]


class WrongLen:

    def __init__(self):
        self.x = x
        self.y = y
        self.batch_size = 2

    def __getitem__(self, idx):
        start_idx, end_idx = idx * self.batch_size, (idx + 1) * self.batch_size
        return self.x[start_idx:end_idx], self.y[start_idx:end_idx]

    def __len__(self):
        return 10  # Should be 5.


class CorrectGenerator(WrongLen):
    def __len__(self):
        return len(self.x) // self.batch_size


class RaiseValueErrorGenerator(OnlyLen):
    def __getitem__(self, idx):
        raise ValueError("From within generator.")


class RaiseTypeErrorGenerator(OnlyLen):
    def __getitem__(self, idx):
        raise TypeError("From within generator.")


class RaiseZeroDivisionErrorGenerator(OnlyLen):
    def __getitem__(self, idx):
        return 1 / 0


class RaiseZeroDivisionErrorWhenIterating(OnlyLen):

    def __getitem__(self, idx):
        if idx == 0:
            return x, y
        return 1 / 0


x = ops.ones((10, 10))
y = ops.ones((10,))


class TestDataHandler:

    @pytest.mark.parametrize(
        "features, labels, raises",
        [
            (x, y, nullcontext()),
            (x, y[:5], pytest.raises(ValueError)),
            (x, None, pytest.raises(TypeError)),
            (CorrectGenerator(), None, nullcontext())
        ]
    )
    def test_constructor(self, features, labels, raises):
        with raises:
            DataHandler(features, labels)

    @pytest.mark.parametrize(
        "generator, expect",
        [
            (OnlyLen, pytest.raises(AttributeError)),
            (OnlyGetitem, pytest.raises(AttributeError)),
            (ReturnThree, pytest.raises(ValueError)),
            (ReturnOneNonIterable, pytest.raises(TypeError)),
            (ReturnOneIterable, pytest.raises(ValueError)),
            (ReturnTwoWrongType, pytest.raises(TypeError)),
            (ReturnDifferentLength, pytest.raises(ValueError)),
            (RaiseValueErrorGenerator, pytest.raises(ValueError)),
            (RaiseTypeErrorGenerator, pytest.raises(TypeError)),
            (RaiseZeroDivisionErrorGenerator, pytest.raises(RuntimeError)),
        ]
    )
    def test_validate_generator(self, generator, expect):
        with expect:
            DataHandler(generator())

    @pytest.mark.parametrize(
        "array, error",
        [
            ([1], nullcontext()),
            ([[1, 2], [2]], pytest.raises(ValueError)),
            (x.astype('d'), nullcontext()),
            (x.astype('float32'), nullcontext()),
            ({'x': 2}, pytest.raises(TypeError))
        ]
    )
    def test_to_ndarray(self, array, error):
        with error:
            assert DataHandler(x, y, 2).to_ndarray(array).dtype == 'float32'

    @pytest.mark.parametrize(
        'test_size, expect', [
            (0.1, nullcontext()),
            (0.01, pytest.raises(ValueError)),
            (0.99, pytest.raises(ValueError)),
            (-1, pytest.raises(ValueError)),
            (0, pytest.raises(ValueError)),
            (1, pytest.raises(ValueError)),
            (1.2, pytest.raises(ValueError))
        ]
    )
    def test_train_test_split(self, test_size, expect):
        with expect:
            DataHandler.train_test_split(x, y, test_size)

    @pytest.mark.parametrize(
        "generator, expect",
        [
            (WrongLen, pytest.raises(IndexError)),
            (RaiseZeroDivisionErrorWhenIterating, pytest.raises(RuntimeError)),
            (CorrectGenerator, nullcontext()),
            (lambda: DataHandler(CorrectGenerator()), nullcontext())
        ]
    )
    def test_getitem_exception_handling(self, generator, expect):
        with expect:
            dataset = DataHandler(generator())
            for idx in range(len(dataset)):
                features, labels = dataset[idx]
                assert len(features) == len(labels)

    @pytest.mark.parametrize(
        "features, labels, batch_size",
        [
            (x, y, 2),
            (CorrectGenerator(), None, 2),
            (CorrectGenerator(), None, None),
            (DataHandler(CorrectGenerator()), None, None),
            (x, y, 1),
            (x, y, None),
            (x, y, len(x)),
            (x, y, 3),
            (x, y, 4),
            (x, y, 11)
        ]
    )
    def test_getitem_return_length(self, features, labels, batch_size):
        dataset = DataHandler(features, labels, batch_size)
        for idx in range(len(dataset)):
            x_batch, y_batch = dataset[idx]
            if batch_size:
                assert len(x_batch) == batch_size or len(x_batch) == len(features) - batch_size * (len(dataset) - 1)
            else:
                if isinstance(features, ops.ndarray):
                    assert len(x_batch) == len(y_batch) == len(features)



"""An example convolution neural network trained on MNIST."""
import hashlib
import os
import shutil
import urllib.error
import urllib.request

import numpy as np

from xrnn import losses
from xrnn.activations import Softmax, ReLU
from xrnn.layers import Dense, Conv2D, Flatten, MaxPooling2D, BatchNormalization
from xrnn.model import Model
from xrnn.optimizers import SGD


def check_mnist_integrity(path: str) -> bool:
    """Checks the mnist file at `path` integrity by verifying its SHA256 checksum."""
    # This is directly stolen from https://github.com/keras-team/keras/blob/master/keras/datasets/mnist.py#L64.
    file_hash = "731c5ac602752760c8e48fbffcf8c3b850d9dc2a2aedcf2cc48468fc17b673d1"  # sha256 hash.
    with open(path, 'rb') as mnist_bytes:
        curr_file_hash = hashlib.sha256(mnist_bytes.read()).hexdigest()
        if curr_file_hash != file_hash:
            print(
                "Hash mismatch, meaning the file is corrupt. This probably happened because the process terminated "
                "or a connection error occurred whilst downloading the dataset. Attempting re-download...")
            return False
        return True


def load_mnist() -> np.ndarray:
    """
    This function loads the mnist dataset if it's already downloaded, if not it downloads it from Google storage api
    first. The installation path is at ~/.xrnn/datasets/mnist.npz (`~` is the user directory`).
    """
    datasets_dir = os.path.join(os.path.expanduser('~'), '.xrnn', 'datasets')
    mnist_path = os.path.join(datasets_dir, 'mnist.npz')
    if os.path.exists(mnist_path):
        if check_mnist_integrity(mnist_path):
            return np.load(mnist_path)
    if not os.path.isdir(datasets_dir):
        os.makedirs(datasets_dir)
    try:
        # This is directly stolen from https://github.com/keras-team/keras/blob/master/keras/datasets/mnist.py#L58.
        url = "https://storage.googleapis.com/tensorflow/tf-keras-datasets/mnist.npz"
        print("Downloading the MNIST dataset...")
        with urllib.request.urlopen(url, timeout=15) as response, open(mnist_path, 'wb') as out_file:  # nosec
            # `nosec` tells `bandit` to ignore security checks for this line, normally bandit considers urlopen as a
            # medium severity vulnerability because it allows use of 'file:' and custom schemes. However, its use here
            # is safe since we verify the downloaded file integrity.
            shutil.copyfileobj(response, out_file)
    except (urllib.error.URLError, urllib.error.ContentTooShortError, TimeoutError) as e:
        if not os.listdir(datasets_dir):  # If it's empty.
            os.rmdir(datasets_dir)
        raise ConnectionError(
            "Couldn't download the data because of a connection error. "
            "Make sure you are connected to the internet and try again.") from e
    if check_mnist_integrity(mnist_path):
        return np.load(mnist_path)
    raise ConnectionError("Dataset file is corrupted, please try downloading it again.")


mnist_dataset = load_mnist()
# The dataset images have the shape (n_samples, height, width), and a convolution layer expects the input to have a
# channels dimension, so we expand the images shape to (n_samples, height, width, 1). 1 means that the images are
# gray scale.
X = np.expand_dims(mnist_dataset['x_train'], 3)
y = mnist_dataset['y_train']

# Do the same to test images as we did to the train images
X_test = np.expand_dims(mnist_dataset['x_test'], 3)
y_test = mnist_dataset['y_test']


def mnist_example(batch_size: int = 128, epochs: int = 1) -> None:
    """
    Train a CNN on the MNIST dataset, validate the network on the test dataset and performs a prediction on a sample of
    data.

    Parameters
    ----------
    batch_size: int, optional
        How many samples each slice of the data should have. A bigger number could result in faster training (up to a
        point) and will also result in more memory consumption.
    epochs: int, optional
        How many full iterations over the dataset to train for.
    """

    # Construct the neural network as a series of layers.
    model = Model()
    model.add(Conv2D(16, 3, 2, 'same'))
    model.add(ReLU())
    model.add(BatchNormalization())
    model.add(MaxPooling2D(2, 2, 'same'))
    model.add(Flatten())
    model.add(Dense(100))
    model.add(ReLU())
    model.add(Dense(10))  # 10 neurons because there are 10 classes in the dataset (digits 0 to 9).
    model.add(Softmax())  # Used to turn the output of the last layer into probabilistic distribution for each class.

    # Build the network with the input shape and batch size.
    model.build((batch_size,) + X.shape[1:])

    # Set the optimization algorithm and loss function.
    model.set(SGD(0.01, 0.9), losses.CategoricalCrossentropy())

    # Print a summary of the network's layers, parameters, input/output shapes and memory consumption.
    model.summary()

    # Train the network on `X` and `y` for `EPOCHS` epochs and validate on `X_test` and `y_test`.
    model.train(X, y, batch_size=batch_size, epochs=epochs, validation_data=[X_test, y_test])

    # Predict the label of the first sample in the test dataset.
    # np.expand_dims is used here to turn the sample (28, 28, 1) into a batch containing one sample (1, 28, 28, 3)
    # because the network works in batches even if it's one sample.
    prediction = model.predict(np.expand_dims(X_test[0], 0))

    # Model predicts on batches, so even if one sample is provided, it's turned into a batch of 1, that's why we take
    # the first sample.
    prediction = prediction[0]

    # The model returns a probability for each label, `np.argmax` returns the index with the largest probability.
    label = np.argmax(prediction)

    print(f"Prediction: {label} - Actual: {y_test[0]}.")


# For when it's run from the command line.
if __name__ == "__main__":
    mnist_example()



"""Defines ReLU, LeakyReLU, Softmax, Tanh and Sigmoid activation functions Layers."""
from xrnn import ops
from xrnn.layers import Layer


class ReLU(Layer):

    def __init__(self, alpha: float = 0.) -> None:
        """
        Rectified linear unit activation function class. This is the most used activation function between network
        layers to introduce non-linearity, but is rarely (if ever) used as an output layer.

        Parameters
        ----------
        alpha: float, optional
            A float representing the negative slope to use. Only used in leaky relu case and is equal to 0 for relu
            according to relu's formula `max(0, inputs)`. Default is 0.
        """
        if not 0 <= alpha < 1:
            raise ValueError('`alpha` must be in [0, 1).')
        super().__init__()
        self.alpha = alpha
        self.relu_where = None

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        # relu = max(0, x)
        relu_where = inputs > 0
        if self.training:
            self.relu_where = relu_where  # Save this result because it's used both in the forward and backward passes
            # for relu and leaky relu. Small memory footprint but measurable performance.
        if not self.alpha:
            output = inputs * relu_where
        else:
            # fills an array with inputs for inputs greater than zero and with inputs * alpha for inputs less than zero.
            output = ops.where(relu_where, inputs, inputs * self.alpha)
        return output

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of relu is: 1(x > 0). And here we apply the chain rule directly (d_relu * gradients).
        if not self.alpha:
            return d_values * self.relu_where
        # derivative of leaky relu is: {1; x > 0, alpha; x < 0)
        dlrelu = ops.ones(self.relu_where.shape)
        dlrelu[~self.relu_where] = self.alpha
        return d_values * dlrelu

    def get_config(self) -> dict:
        activation_config = super().get_config()
        activation_config.update({'alpha': self.alpha})
        return activation_config


class LeakyReLU(ReLU):

    def __init__(self, alpha: float = 0.01) -> None:
        """
        Leaky ReLU activation function class. Leaky relu is used to solve relu's dying problem by introducing a slight
        slope in the negative range which cases the output of it have small negative numbers below zero.

        `LeakyReLU(x) = alpha * x, x < 0`

        `LeakyReLU(x) = x, x > 0`

        Parameters
        ----------
        alpha: float, optional
            The negative slope coefficient. Default is 0.01
        """
        super().__init__(alpha)


class Sigmoid(Layer):
    """
    Sigmoid activation function layer. It's also called a squashing function because its range is (0, 1). Meaning the
    output of a forward pass through sigmoid is going to result in an array having values between 0 and 1 no matter
    inputs. This layer is usually as the output layer of binary classifier because its output is in the range of (0, 1)
    so for example, values < 0.5 can be classified as class 1 and values greater than 0.5 as class 2.
    """

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        # sigmoid = 1 / (1 + e^-z)
        output = 1 / (1 + ops.exp(-inputs))
        if self.training:
            self.output = output
        return output

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of sigmoid is: sigmoid * (1 - sigmoid)
        # We have to calculate the sigmoid in parentheses because we need to calculate first then multiply it by the
        # inputs from the next layer to apply the chain rule. Or the order of operations can be changed
        # Sigmoid * (1 - Sigmoid) * d_values. Any other order of operations (without parentheses) is incorrect.
        return d_values * (self.output * (1 - self.output))


class Softmax(Layer):
    """
    Softmax activation function layer. It's usually used as the final layer of classifier network because it converts
    the inputs into a probability distribution of K possible outcomes, where K is the number of classes/labels.
    For example, when building a CNN classifier that takes images of cats, dogs or cars and classifies them, the last
    two layers will be:
    Dense(3)  # Because there are three classes (dog, cat, car)
    Softmax()  # To turn the output to a probabilistic distribution.
    """

    def __init__(self) -> None:
        super().__init__()
        self.expected_dims = 2

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        # softmax = e^zi / sum(e^zj)
        # Subtract the biggest value in the row from the row values
        exp_values = ops.exp(inputs - ops.amax(inputs, axis=1, keepdims=True))
        # Normalize them for each sample. Divide each value in each row by the sum of that row
        probabilities = exp_values / ops.sum(exp_values, axis=1, keepdims=True)
        if self.training:
            self.output = probabilities
        return probabilities.copy()

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of softmax is: Sij * Kronecker delta - Sij * Sik

        # Note that this is slow, and is rarely used because softmax is usually used with categorical cross entropy and
        # a fast path is used for that instead of calling loss.backward() then softmax.backward().

        # Create uninitialized array
        d_inputs = ops.empty_like(d_values)  # Derivative with respect to inputs.
        # Enumerate outputs and gradients
        for index, (single_output, single_dvalues) in enumerate(zip(self.output, d_values)):
            # Flatten output array so we can use a dot product
            single_output = single_output.reshape(-1, 1)

            # The first part of the derivative. Which equals 1 * softmax output for the ith index and 0 everywhere else.
            kronecker_array = ops.diagflat(single_output)

            # The second part of the derivative (Sij * Sik). Calculate the impact of each value in the array on every
            # other value.
            combinations = ops.dot(single_output, single_output.T)

            # Calculate Jacobian matrix of the output
            jacobian_matrix = kronecker_array - combinations
            # Each row of the Jacobin matrix represents the impact of each input ith on all the other inputs

            # Calculate sample-wise gradient
            # and add it to the array of sample gradients
            d_inputs[index] = ops.dot(jacobian_matrix, single_dvalues)  # Multiply each row in the jacobin matrix
            # by the loss gradient vector. Since the gradient vector is zeros except for the kth index, we end up
            # only calculating the impact of each ith input on the kth input
        return d_inputs


class Tanh(Layer):
    """
    Tanh activation function layer. The output values of tanh are between (-1, 1). One of the uses of tanh is as an
    output layer/activation for generative image models like DCGAN [1]_.

    References
    ----------
    .. [1] https://www.tensorflow.org/tutorials/generative/dcgan#the_generator
    """

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        # Tanh = e^zi - e^-zi / e^zi + e^-zi
        output = ops.tanh(inputs)
        if self.training:
            self.output = output
        return output

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        # The derivative of tanh is: 1 - tanh(z)^2
        return d_values * (1 - ops.square(self.output))



import pytest

from xrnn import metrics
from xrnn import ops

BATCH_SIZE = 64


def one_hot_encoded(n_classes):
    integers = ops.random.integers(0, n_classes, (BATCH_SIZE,))
    one_hot = ops.zeros((BATCH_SIZE, n_classes))
    one_hot[ops.arange(len(integers)), integers] = 1
    return one_hot


class TestAccuracy:

    def test_init(self):
        metrics.Accuracy('mse')
        metrics.Accuracy('Categorical_crossentropy')
        metrics.Accuracy('bInary_crossentropy')
        with pytest.raises(ValueError):
            metrics.Accuracy('mean_squared_error')

    @pytest.mark.parametrize(
        "y_true, loss",
        [
            # The same test cases as the losses'.
            (ops.random.integers(0, 10, (BATCH_SIZE, 1)), 'categorical_crossentropy'),
            (ops.random.integers(0, 10, (BATCH_SIZE,)), 'categorical_crossentropy'),
            (one_hot_encoded(10), 'categorical_crossentropy'),
            (ops.random.integers(0, 2, (BATCH_SIZE,)), 'binary_crossentropy'),
            (ops.random.integers(0, 2, (BATCH_SIZE, 1)), 'binary_crossentropy'),
            (one_hot_encoded(1), 'binary_crossentropy'),
            (one_hot_encoded(2), 'binary_crossentropy'),
            (one_hot_encoded(3), 'binary_crossentropy'),
            (ops.random.random((BATCH_SIZE, 1)), 'mse'),
            (ops.random.random((BATCH_SIZE,)), 'mse'),
            (ops.random.uniform(-1, 1, (BATCH_SIZE, 16, 16, 3)), 'mse'),
            (ops.random.standard_normal((BATCH_SIZE, 2)), 'mse'),
            (ops.random.integers(0, 2768, (BATCH_SIZE, 100)), 'mse'),
        ]
    )
    def test_calculate(self, y_true, loss):
        acc = metrics.Accuracy(loss)
        y_pred = y_true
        if loss == 'categorical_crossentropy':
            # Simulate the output of a dense/softmax layer with 10 neurons.
            y_pred = ops.random.uniform(0, 1, (BATCH_SIZE, 10))
            if y_true.ndim == 2:
                if y_true.shape[1] != 1:  # one-hot encoded
                    y_true_indices = ops.argmax(y_true, 1)
                else:  # Just an extra dim.
                    y_true_indices = ops.squeeze(y_true.copy())
            else:
                y_true_indices = y_true  # sparse (integer) labels
            y_pred[ops.arange(len(y_true)), y_true_indices] = 1  # match one-hot y_pred to y_true to get 100% accuracy.
        assert acc(y_true, y_pred) == 1.



"""
`pyproject.toml` isn't sufficient and `setup.py` is needed for two reasons:
  1. setuptools versions available for Python 3.6 don't support `pyproject.toml`, so the package metadata needs to be
     present in `setup.cfg` or `setup.py` when building the package using Python 3.6.
  2. Custom build commands are used for building the package, and there's no way to achieve that using `pyproject.toml`.

Defining the package metadata in two places is sure redundant, but causes no conflicts, because when using a setuptools
version that supports `pyproject.toml`, its metadata overrides `setup.py` metadata.
"""
import os
import ast
from pathlib import Path
import platform
from typing import Tuple, List
import subprocess

from setuptools import setup
from setuptools.dist import Distribution
from setuptools._distutils.sysconfig import customize_compiler
from setuptools._distutils.ccompiler import new_compiler
# Can't import distutils directly because it's deprecated and was removed in python 3.12.
from wheel.bdist_wheel import bdist_wheel


def is_apple_silicon_native() -> bool:
    """Returns True if Python is running on Apple Silicon natively."""
    # This only returns 'arm' when running natively on Apple Silicon, and not under Rosetta, which is desired behaviour.
    return platform.system() == 'Darwin' and platform.processor() == 'arm'


def is_clang() -> bool:
    """Returns whether the compiler being used is Clang."""
    # If clang is explicitly set as the default compiler.
    if os.environ.get('CC', None):
        return 'clang' in os.environ['CC']
    # When no compiler is explicitly set, `cc` is used, we check if it's clang by checking the output of `-v`.
    # Weirdly, the output of `cc -v` is redirected to stderr not stdout.
    return b'clang' in subprocess.run(['cc', '-v'], stderr=subprocess.PIPE).stderr.lower()


def get_compiler():
    """
    Returns an instance of CCompiler that can be used to compile C/C++ source code that uses the default compiler
    on each platform (MSVC on Windows, GCC on Linux, Clang on MacOS).
    The compiler type can be defined in `pyproject.toml` under the key `compiler`.
    """
    with open('pyproject.toml', 'r') as config_file:
        for line in config_file.readlines():
            line = "".join(line.split())
            idx = line.find('compiler=')
            if idx != -1:
                compiler_type = ast.literal_eval(line[idx + len('compiler='):])
    if compiler_type:
        os.environ['CC'] = compiler_type
    if platform.system() == 'Windows':
        if any(comp_id in compiler_type for comp_id in ['gcc', 'clang', 'mingw']):
            compiler_type = 'mingw32'
        compiler = new_compiler(compiler=compiler_type or None)
        compiler.shared_lib_extension = '.dll'  # Default dll extension for GCC and Clang on Windows is 'dll.a'.
    else:
        compiler = new_compiler()
        if compiler_type:
            customize_compiler(compiler)  # Need to call this function to use the `CC` defined above.
    return compiler


def get_compiler_flags(compiler) -> Tuple[List[str], List[str]]:
    """Return a tuple containing compiler and linker flags based on the compiler and platform."""
    if 'msvc' in type(compiler).__name__.lower():
        # /MT: Link the CRT statically, so we don't have to ship it alongside the package.
        compiler_flags = ['/openmp', '/Ox', '/MT']
        linker_flags = []
    elif platform.system() == 'Windows':  # For GCC (MinGW) or Clang on Windows.
        compiler_flags = ['-fopenmp', '-O3']  # Need to link statically because there's no automatic/easy way
        # to bundle openmp library with the package on Windows, unlike on Linux and macOS.
        linker_flags = ['-fopenmp', '-lucrt', '-static']
        compiler.dll_libraries = []
        # Empty this list because the code doesn't depend on msvcr and there's no need to include it.
    elif platform.system() == 'Darwin':
        libomp_dir_path = "/opt/homebrew/opt/libomp" if is_apple_silicon_native() else "/usr/local/opt/libomp"
        # /usr/local/opt/ is the installation path on Intel and Rosetta 2.
        compiler_flags = [f"-I{libomp_dir_path}/include", "-fopenmp", "-O3", "-fPIC"]
        linker_flags = [f"-L{libomp_dir_path}/lib", "-lomp"]
        if is_clang():
            compiler_flags.insert(1, '-Xclang')
    else:
        compiler_flags = ['-fopenmp', '-O3', '-fPIC']
        linker_flags = ['-fopenmp']
    return compiler_flags, linker_flags


class ExtModules(list):
    """Empty list that tests as truthful."""
    def __bool__(self) -> bool:
        return True


class BinaryDistribution(Distribution):
    """
    Custom Distribution which tells that it has C extension modules. This is needed because the package does have source
    files that it builds a shared library from, they just don't depend on the Python C API, and we use a custom build
    method to build the shared library. This way when creating a wheel, it's created like we have a distribution with
    C extension (a platform lib) which is what we want.
    """

    def __init__(self, *args, **kwargs):
        Distribution.__init__(self, *args, **kwargs)
        self.cmdclass['bdist_wheel'] = BuildSharedLib  # To call our custom build step when building the wheel.
        self.ext_modules = ExtModules()  # List that tests as truthful to trick the distribution to build a platlib.
        # See https://github.com/microsoft/debugpy/blob/main/setup.py#L36-L46 for more information about why it's needed

    def has_ext_modules(self) -> bool:
        """To trick the distribution to think it has C extension modules."""
        return True


class BuildSharedLib(bdist_wheel):
    """
    Custom build command for creating the built distribution (wheel) from the source distribution.
    See `get_tag()` and `run()` for more information about how it works.
    """

    def get_tag(self) -> Tuple[str, str, str]:
        """
        The build is only platform-dependent, so we only get the platform tag from the full tag (containing Python
        version and ABI version), and replace Python and ABI to 'py3' and 'none' respectively.
        We do that because the built shared library is independent of the Python API, it's not an extension module, and
        by doing that we'll only need to build the wheel once and not for every Python version.
        """
        python, abi, platform_tag = bdist_wheel.get_tag(self)
        return 'py3', 'none', platform_tag

    def run(self) -> None:
        """
        Custom `run` function to compile source files into a shared library from,
        place it in the build directory (lib/), and include it in the built distribution (wheel).
        """
        # Create an instance of CCompiler (msvc, gcc, cygwin, etc)
        compiler = get_compiler()
        compiler_flags, linker_flags = get_compiler_flags(compiler)

        # Prepare paths.
        shared_lib_filename = 'c_layers' + compiler.shared_lib_extension  # The name with extension.
        sources_paths = [os.path.join('xrnn', source_filename) for source_filename in ['layers_f.c', 'layers_d.c']]
        build_dir = os.path.join('xrnn', 'lib')  # Where the built library would reside.
        shared_lib_relative_path = os.path.join('lib', shared_lib_filename)  # This is relative from the package source.

        # Need to manually create the output directory for the shared library or a link error might occur.
        if not os.path.exists('lib'):
            os.mkdir('lib')

        # Compile the source into object files.
        objects = compiler.compile(sources_paths, extra_postargs=compiler_flags)

        # Create a dynamic/shared library from the object files named `shared_lib_filename` and place at `build_dir`.
        compiler.link_shared_object(objects, shared_lib_filename, build_dir, extra_postargs=linker_flags)

        # Add the library to the package data, so it would be included in the package's directory.
        self.distribution.package_data['xrnn'].append(shared_lib_relative_path)

        bdist_wheel.run(self)


# To read the README.md file and use it to populate `long_description`. Basically what readme="README.md" do.
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name="xrnn",
    version="1.1.1",
    author="Yazan Sharaya",  # This is done in one line in `pyproject.toml`: authors = [{name = "", email = ""}]
    author_email="yazan.sharaya.yes@gmail.com",
    description="Light weight fast machine learning framework.",
    long_description=long_description,  # These two lines are equivalent to: readme = "README.md" in `pyproject.toml`.
    long_description_content_type="text/markdown",
    license="MIT",
    python_requires=">=3.6",
    install_requires=[  # Same as dependencies in `pyproject.toml`.
        "numpy>=1.17",
        "typing-extensions; python_version<'3.8'",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: C",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development"
    ],

    extras_require={  # Same as [project.optional-dependencies] in `pyproject.toml`.
        "test": [
            "pytest"
        ],
        "dev": [
            "build"
        ]
    },

    project_urls={  # Same as [project.urls]; repository = "" in `pyproject.toml`.
        "Repository": "https://github.com/Yazan-Sharaya/xrnn",
    },

    license_files=("LICENSE", ),  # Same as [tool.setuptools]; license-files = ["LICENSE"] in `pyproject.toml`.
    packages=["xrnn"],  # Same as [tool.setuptools]; packages = ["xrnn"] in `pyproject.toml`.

    package_data={"xrnn": ["*.c"]},  # Same as [tool.setuptools.package-data] in `pyproject.toml`.
    include_package_data=True,  # This line isn't present in `pyproject.toml` because it's implicitly set to True there.

    distclass=BinaryDistribution,  # A custom distribution class that is needed because of the custom build process.
)
